package botworld;

import java.awt.*;
import java.io.*;
import java.util.Enumeration;

import prolog.*;
import prolog.math.*;

/**
 * Bots.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Bot.java,v 1.11 2000/08/10 12:51:03 peteg Exp $
 */
public class Bot extends WorldObject implements Runnable {
    /* Dimensions of a bot. */
    public static final int arm_length = 12;
    public static final int arm_angle = 30;
    public static final int hand_radius = 3;
    public static final int radius = 14;

    /**
     * The Prolog engine.
     */
    Prolog prolog;

    /**
     * The angle the bot is facing.
     */
    float angle;

    /** What we're holding, if anything. */
    Bar held;

    /**
     * Our parent - used to get information about the world.
     */
    Botworld parent;

    /**
     * What the bot is saying.
     */
    String speech;

    /**
     * Creates a new bot, with specifying all parameters.
     */
    public Bot(int bid, int centreX, int centreY, float init_angle,
	       Database db, Botworld botworld) {

	try {
	    id = bid;
	    bounds = new Rectangle(centreX - radius,
				   centreY - radius,
				   2 * radius,
				   2 * radius);
	    angle = init_angle;
	    prolog = new Prolog(db);
	    prolog.obj = this;
	    held = null;
	    parent = botworld;
	    speech = "";
	} catch(Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * Creates a new bot, partially specified.
     */
    Bot(int bid, Database db, Botworld botworld, boolean time) {
	/* Normal init. */
	this(bid, 0, 0, 0, db, botworld);

	/* World bounds. */
	double x_scale = botworld.dim.width - radius,
	    y_scale = botworld.dim.height - radius;

	/* Find an initial position and angle for the bot. */
	do {
	    bounds.x = (int)(Math.random() * x_scale);
	    bounds.y = (int)(Math.random() * y_scale);
	} while(parent.collided(this));

	angle = (float)(Math.random() * 359);
	prolog.time = time;
    }

    /**
     * The main thread of execution.
     */
    public void run() {
	/* Put the static objects in the Prolog database. */
	int size = parent.staticObjects.size();

	for(int i = 0; i < size; i++) {
	    WorldObject wo = (WorldObject)parent.staticObjects.elementAt(i);
//  	    System.err.println("run #" + id + " " + wo.getFact());
 	    prolog.db.put(new Clause(wo.getFact(), prolog));
	}

	/* Execute the bot's code. */
 	Compound arg0 = new Compound("start", new Term[]{IntNum.make(id)});
 	Goal query = prolog.toGoal(arg0);

	try {
	    prolog.once(query);
	} catch(PrologException e) {
	    e.printStackTrace();
	}

	System.err.println("Bot #" + id + " finished.");
    }

    /**
     * Looks at the world.
     *
     * @returns an enumeration of the movable objects in the world.
     */
    public Enumeration look() {
	return(parent.movableObjects.elements());
    }

    /**
     * Moves the bot to (x,y). Rotate, calc dist, move forward.
     */
    public void move(float x, float y) {
// 	System.err.println("move: (" + x + ", " + y + ")");
	float dx = x - (bounds.x + radius),
	    dy = y - (bounds.y + radius);

	turnTo(x, y);
	move(Math.round((float)Math.sqrt(dx * dx + dy * dy)));
    }

    /**
     * Moves the bot forward x units in the direction it's currently facing.
     */
    public void move(float dist) {
// 	System.err.println("move: dist = " + dist);

	float dx = (float)Math.cos(angle * RADIANS),
	    dy = (float)Math.sin(angle * RADIANS),
	    x = bounds.x,
	    y = bounds.y;

	/* Distance of held object from our centre. */
	float held_dist = (radius + arm_length) * (float)Math.cos(arm_angle * RADIANS);

    move: while(dist-- > 0) {
	bounds.x = (int)(x += dx);
	bounds.y = (int)(y += dy);

	/* Move what we're holding as well. Recompute it relative to the bot
	 * to minimise rounding errors.
	 */
	if(held != null) {
	    float bot_x = x + radius,
		bot_y = y + radius;

	    held.bounds.x = (int)(bot_x + held_dist * Math.cos(angle * RADIANS));
	    held.bounds.y = (int)(bot_y + held_dist * Math.sin(angle * RADIANS));
	}

	if(parent.collided(this)) {
	    System.err.println("Bot(" + id + ") collided!");
	    bounds.x = (int)(x -= dx);
	    bounds.y = (int)(y -= dy);
	    break move;
	}

	parent.repaint();
 	try {
 	    Thread.sleep(10);
 	} catch (InterruptedException e) {}
    }
    }

    /**
     * Turns the bot to face (x, y).
     */
    public void turnTo(float x, float y) {
	int dx = Math.round(x - (bounds.x + radius)),
	    dy = Math.round(y - (bounds.y + radius));
	double theta;

	if(dx == 0 && dy == 0)
	    return;

	theta = (Math.atan2(dy, dx) * DEGREES - angle) % 360;

	/* Choose the direction to rotate that's the smallest. */
	if(theta >= 0) {
// 	    System.err.println("turnTo: theta = " + theta + " 360 - theta = " + (360 - theta));
	    turn((theta > 360 - theta) ? 360 - theta : theta);
	} else {
// 	    System.err.println("turnTo: theta = " + theta + " 360 + theta = " + (360 + theta));
	    turn((-theta > 360 + theta) ? 360 + theta : theta);
	}
    }

    /**
     * Rotates the bot x degrees.
     */
    public void turn(double degs) {
// 	System.err.println("turn " + degs + " degrees - angle = " + angle);

	int mag = (int)(degs >= 0 ? degs : -degs);
	double sign = degs >= 0 ? 1 : -1;

	/* Distance of held object from our centre. */
	double held_dist = (radius + arm_length) * Math.cos(arm_angle * RADIANS),
	    bot_x = bounds.x + radius,
	    bot_y = bounds.y + radius;

	while(mag-- > 0) {
	    angle += sign;
	    if(angle < 0.0)
		angle = (float)359.0;
	    else
		angle %= (float)360.0;

// 	    System.err.println(" mag = " + mag + " angle = " + angle);

	    /* Move what we're holding as well. Recompute it relative to the
	     * bot to minimise rounding errors.
	     */
	    if(held != null) {
		held.bounds.x = (int)(bot_x + held_dist * Math.cos(angle * RADIANS));
		held.bounds.y = (int)(bot_y + held_dist * Math.sin(angle * RADIANS));
		held.angle = (angle + 90) % 360;
	    }

	    parent.repaint();

 	    try {
 		Thread.sleep(10);
 	    } catch (InterruptedException e) {}
	}
    }

    /**
     * Grabs a bar. bar.onbar() better be true first, tho.
     */
    public void grab(int barid) {
// 	System.err.println("grab: " + barid);
	if(held != null)
	    return;

	/* Linear search for the bar id (ouch). */
	Bar bar = null;

	for(Enumeration e = parent.bars.elements();
	    e.hasMoreElements(); ) {

	    bar = (Bar)e.nextElement();
	    if(bar.id == barid)
		break;
	}

 	if(bar != null && bar.onBar(this) && bar.lock()) {
	    System.err.println(" got it.");
 	    held = bar;
	}
    }

    /**
     * Drops a bar, if we're holding one.
     */
    public void drop() {
	if(held != null) {
	    held.unlock();
	    held = null;
	}
    }

    /**
     * Says something.
     */
    public void say(String s) {
	speech = (s != null ? s : "");
	parent.repaint();
    }

    /**
     * Detaches the bot.
     */
    public void detach() {
    }

    public Compound getFact() {
	Term[] args = new Term[] {
	    IntNum.make(id),
	    IntNum.make(bounds.x + radius),
	    IntNum.make(bounds.y + radius),
	    new FloNum(Math.round(angle)),
	    IntNum.make((held == null) ? 0 : held.id),
	    new Compound(speech),
	};

	return(new Compound("bot", args));
    }

    public void draw(Graphics g) {
	g.setColor(Color.white);
	g.fillOval(bounds.x, bounds.y, 2 * radius, 2 * radius);

	/* Draw the arms. */
	int x = bounds.x + radius,
	    y = bounds.y + radius,
	    leftx = (int)(Math.cos((angle - arm_angle) * RADIANS) * (radius + arm_length) + x),
	    lefty = (int)(Math.sin((angle - arm_angle) * RADIANS) * (radius + arm_length) + y),
	    rightx = (int)(Math.cos((angle + arm_angle) * RADIANS) * (radius + arm_length) + x),
	    righty = (int)(Math.sin((angle + arm_angle) * RADIANS) * (radius + arm_length) + y);

	g.drawLine(x, y, leftx, lefty);
	g.drawLine(x, y, rightx, righty);

	/* ...and the hands. */
	g.drawOval(leftx - hand_radius, lefty - hand_radius, hand_radius * 2, hand_radius * 2);
	g.drawOval(rightx - hand_radius, righty - hand_radius, hand_radius * 2, hand_radius * 2);

	/* Add the speech. */
	if(speech != null)
	    g.drawString(speech, bounds.x, bounds.y - 15);
    }
}
