package prolog;

import java.util.Enumeration;

import prolog.util.Hashtable;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * A compiled bucket from a <TT>ClauseIndex</TT>.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: ClauseIndexTrad.java,v 1.18 2000/11/13 11:55:04 peteg Exp $
 */
class CCGBucket {
    CCGBucket next;

    /**
     * The thing we're indexing on. -1 == default case.
     */
    int index;

    /**
     * The start of the instruction sequence for this bucket.
     */
    InstructionHandle ih;

    protected ClauseCodeGen first;
    protected ClauseCodeGen last;

    CCGBucket(InstructionHandle ih) {
	this.ih = ih;
    }

    /**
     * Adds a clause to the end of the predicate.
     */
    protected void putz(ClauseCodeGen ccg) {
	if(last == null) {
	    first = last = ccg;
	} else {
	    last = last.next = ccg;
	}
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	for(ClauseCodeGen ccg = first; ccg != null; ccg = ccg.next) {
	    out.append(ccg);
 	    out.append("\n");
	}

	if(next != null) {
	    out.append(next.toString());
	}

 	return out.toString();
    }
}

/**
 * A traditional clause index: consider only the functor of the first
 * argument of the clauses and goal. Note that we assume the arity of the
 * predicate is non-zero.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: ClauseIndexTrad.java,v 1.18 2000/11/13 11:55:04 peteg Exp $
 */
class ClauseIndexTrad extends ClauseIndex {
    /**
     * A hash, keyed by the functor of the first argument of the goal. I
     * assume there won't be many distinct functors.
     * <P>
     * Maps Numeric/FunctorArity -> ClauseIndexRec
     */
    protected Hashtable functors;
 
    /**
     * A <TT>ClauseList</TT> of clauses with variables as their first
     * arguments - i.e. those that must be in all buckets.
     */
    private ClauseIndexRec variables;
 
    ClauseIndexTrad() {
	super();
	functors  = new Hashtable();
	variables = new ClauseIndexRec();
    }

    protected ClauseList getClauses(Term[] args) {
	Term t = args[0].getBinding();

	if(t instanceof Binding) {
	    // Not much we can do.
	    return super.getClauses(args);
	} else if (t instanceof Compound) {
	    // Return the appropriate bucket, or null if there isn't one.
	    ClauseIndexRec cir = (ClauseIndexRec)functors.get(((Compound)t).fa);

	    return (cir == null ? variables : cir).firstClause;
	} else {
	    // Must be a Num, so do the same as for a Compound.
 	    ClauseIndexRec cir = (ClauseIndexRec)functors.get(t);

	    return (cir == null ? variables : cir).firstClause;
	}
    }

    protected void puta(Clause c) {
	super.puta(c);

	Term t = c.head.args[0];

	if(t instanceof Variable) {
	    // Add to the "variables" list, and all buckets.
	    variables.puta(c);

	    for(Enumeration e = functors.keys(); e.hasMoreElements(); ) {
		((ClauseIndexRec)e.nextElement()).puta(c);
	    }
	} else if (t instanceof Compound) {
	    // Add to the appropriate bucket.
	    get(((Compound)t).fa).puta(c);
	} else {
	    // Must be a Num, so do the same as for a Compound.
 	    get(t).puta(c);
	}
    }

    protected void putz(Clause c) {
// 	System.err.println("CIT.putz(): " + c);
	super.putz(c);

	Term t = c.head.args[0];

	if(t instanceof Variable) {
// 	    System.err.println(">>  CIT.putz() variables");
	    // Add to the "variables" list, and all buckets.
	    variables.putz(c);

	    for(Enumeration e = functors.elements(); e.hasMoreElements(); ) {
		((ClauseIndexRec)e.nextElement()).putz(c);
	    }
	} else if (t instanceof Compound) {
	    // Add to the appropriate bucket.
	    get(((Compound)t).fa).putz(c);
	} else {
	    // Must be a Num, so do the same as for a Compound.
 	    get(t).putz(c);
	}
    }

    public void clear() {
	super.clear();
	functors.clear();
	variables.clear();
    }

    void remove(Clause c) {
	super.remove(c);

	Term t = c.head.args[0];

	if(t instanceof Variable) {
	    // Remove from the "variables" list, and all buckets.
	    variables.remove(c);

	    for(Enumeration e = functors.elements(); e.hasMoreElements(); ) {
		((ClauseIndexRec)e.nextElement()).remove(c);
	    }
	} else if (t instanceof Compound) {
	    // Remove from the appropriate bucket.
	    ((ClauseIndexRec)functors.get(((Compound)t).fa)).remove(c);
	} else {
	    // Must be a Num, so do the same as for a Compound.
 	    ((ClauseIndexRec)functors.get(t)).remove(c);
	}
    }

    private ClauseIndexRec get(Object o) {
	ClauseIndexRec cir;

	if((cir = (ClauseIndexRec)functors.get(o)) == null) {
	    functors.put(o, cir = (ClauseIndexRec)variables.clone());
	}

	return cir;
    }

    void compileClauses(PredicateCodeGen pcg, TermInfo[] arg_tis) {
	// Consider the type of the first argument.
	TermInfo arg0_ti = arg_tis[0];

	if(arg0_ti.type instanceof TypeVariable) {
	    // Not much we can do.
	    super.compileClauses(pcg, arg_tis);
	} else if (arg0_ti.type instanceof NumericType) {
	    // A number. FIXME: treat it like a variable for the time being.
	    super.compileClauses(pcg, arg_tis);
// 	    throw new RuntimeException();
	} else if(arg0_ti.type instanceof CompoundType) {
	    // Only one bucket needs looking at.
	    throw new RuntimeException();
	} else if(arg0_ti.type instanceof AlgebraicType) {
	    algebraic_build_index(pcg, (AlgebraicType)arg0_ti.type, arg_tis);
	} else {
	    // Unknown type.
	    throw new RuntimeException();
	}
    }

    void algebraic_build_index(PredicateCodeGen pcg,
			       AlgebraicType at,
			       TermInfo[] arg_tis) {
// 	System.err.println(">> CIT.alg_build_index(): " + this);

	int num_components = at.components.length;

	// The list of CCG's + where they start, and how many we've got.
	CCGBucket first = null, last = null, vars_ccgb = null;
	int ccg_count = 0;

	// The instruction list from the PCG
	InstructionList il = pcg.il;
	InstructionFactory inst_fact = pcg.inst_fact;

	// Grotty: Add 2 nops. First == start of code, Second == start of
	// first clause. The index_start thing is set to the second, as
	// insert_il() does it before a given ih, not after.
	il.append(inst_fact.NOP);
	InstructionHandle index_start = il.append(inst_fact.NOP);
	LocalVariableGen arg0_lv = new LocalVariableGen(JITDefs.first_free_lv - 1,
							"arg0_lv",
							JITDefs.term_array_type,
							null,
							null);

	for(int i = 0; i < num_components; i++) {
	    CompoundType component = at.components[i];
	    ClauseIndexRec cir = (ClauseIndexRec)functors.get(component.fa);

// 	    System.err.println("CIT: algebraic type: component: "
// 			       + component + " -> "
// 			       + cir);

	    // If there's a bucket for this...
	    CCGBucket ccgb = new CCGBucket(il.getEnd());

	    if(cir != null) {
		// ...compile it.
		ccgb.index = component.functor_arity;

		// Compile the clauses in the bucket.
		for(ClauseList cl = cir.firstClause;
		    cl != null;
		    cl = cl.next) {

		    // FIXME: change type -> component. see CI.set()
		    arg_tis[0].type = component;
		    arg_tis[0].location = new Location(arg0_lv, null);

		    ClauseCodeGen ccg = pcg.compileClause(cl.clause,
						      cl.next == null);

		    if(ccg != null) {
			ccgb.putz(ccg);
		    }
		}

		// If any of that succeeded tack this CCGB onto the list.
		if(ccgb.index != -1 && ccgb.first != null) {
// 		    System.err.println(" >> Adding CCGB to the list.");
		    if(first == null) {
			first = last = ccgb;
		    } else {
			last = last.next = ccgb;
		    }

		    ccg_count++;
		}

		// Reset the type.
		arg_tis[0].type = at;
		arg_tis[0].location = null;
	    } else {
		// ...otherwise use the vars bucket.
		if(variables == null || vars_ccgb != null) {
// 		    System.err.println(" >> no vars bucket or already compiled.");
		    continue;
		} else {
// 		    System.err.println(" >> Compiling the vars bucket: " + variables);
		    cir = variables;

 		    vars_ccgb = ccgb;
		    ccgb.index = -1;

		    // Compile the clauses in the bucket.
		    for(ClauseList cl = cir.firstClause;
			cl != null;
			cl = cl.next) {

			ClauseCodeGen ccg = pcg.compileClause(cl.clause,
							      cl.next == null);

			if(ccg != null) {
			    ccgb.putz(ccg);
			}
		    }
		}
	    }
	}

	// If none of the clauses succeeded, we fail.
	InstructionHandle fail = pcg.failure();

	// If we've got no vars and the indexer doesn't work, we fail.
	InstructionHandle default_target = (vars_ccgb == null)
	    ? fail
	    : vars_ccgb.ih;

// 	System.err.println("\n>> CIT:\n " + il);

	// Create the index.

	// If we at least 2 buckets...
	if(ccg_count > 1 || (vars_ccgb != null && ccg_count == 1)) {
// 	    System.err.println(" >> Adding index.");

	    // 1. Load ((Compound)args[0]).functor_arity
	    InstructionList index_il = new InstructionList();

	    index_il.append(JITDefs.args_load_inst);
	    index_il.append(inst_fact.ICONST_0);
	    index_il.append(inst_fact.AALOAD);
	    index_il.append(inst_fact.createInvoke(TermType.ClassName,
						   JITDefs.term_getBinding,
						   JITDefs.term_getBinding_ret_type,
						   JITDefs.term_getBinding_args_type,
						   JITDefs.INVOKEVIRTUAL));
	    index_il.append(inst_fact.createCheckCast(CompoundType.JavaType));
	    index_il.append(inst_fact.DUP);
	    index_il.append(new ASTORE(arg0_lv.getIndex()));
	    index_il.append(inst_fact.createGetField(CompoundType.ClassName,
						     JITDefs.compound_functor_arity_name,
						     JITDefs.compound_functor_arity_type));

	    // 2. Use an if/else if we've got 2 choices, otherwise switch()
	    if(ccg_count <= 2) {
// 		System.err.println(" >> if/else: if: " + first.ih + " second: " + first.next.ih);

		index_il.append(new PUSH(pcg.cp, first.index));

		if(vars_ccgb == null) {
		    InstructionHandle second_ih = first.next.ih;

		    index_il.append(new IF_ICMPNE(second_ih));
		} else {
		    index_il.append(new IF_ICMPNE(vars_ccgb.ih));
		}
	    } else {
		int[] match                 = new int[ccg_count];
		InstructionHandle[] targets = new InstructionHandle[ccg_count];

		for(int i = 0; first != null; first = first.next, i++) {
		    match[i]   = first.index;
		    targets[i] = first.ih;

// 		    System.err.println(">> CIT: " + match[i] + " -> " + targets[i]);
		}

		index_il.append(new SWITCH(match,
					   targets,
					   default_target,
					   2));
	    }

	    // Wack the indexer in at the start.
	    il.insert(index_start, index_il);

// 	} else if(vars_ccgb != null) {
//  	    throw new RuntimeException();
	}

	// Patch it back to the original type.
	arg_tis[0].type = at;
    }

    public String toString() {
	StringBuffer out = new StringBuffer("ClauseIndexTrad:\n");

	out.append("Variables:\n");
	out.append(variables);
	out.append("\n");

	for(Enumeration e = functors.keys(); e.hasMoreElements(); ) {
	    Object o = e.nextElement();

	    out.append("ClauseIndexRec for ");
	    out.append(o);
	    out.append(":\n");
	    out.append(functors.get(o));
	    out.append("\n");
	}

 	return out.toString();
    }
}
