package prolog;

import prolog.lib.unify_2;

import prolog.util.Vector;

/**
 * <TT>','/2</TT>: Represents a conjuctive goal in the body of a clause.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Conjunction.java,v 1.24 2000/11/12 12:22:44 peteg Exp $
 */
public class Conjunction extends Goal {
    public static final String FUNCTOR = ",".intern();
    public static final int    ARITY   = 2;

    Compound g;
    Goal next;

    /**
     * The predicate record of a possible unifier for this goal.
     */
    Predicate unifier;

    public Conjunction(Compound g, Goal next, Prolog p) {
	this.g    = g;
	this.next = next;

	if((this.unifier = p.db.get(g)) == null) {
	    /* Create an empty, nothing-special Predicate record for this
	     * goal.  */
	    p.db.put(g.functor, g.arity, unifier = p.db.newPredicate(g.arity));
	}
    }

    /**
     * For use by normalise
     *
     * @see #normalise
     */
    Conjunction(Compound g, Predicate unifier) {
	this.g       = g;
	this.unifier = unifier;
    }

    public boolean unify(Term[] insts, Prolog p) throws PrologException {
	p.set_pending_goal(next);

	return unifier.unify(g.instantiateHead(insts), p);
    }

    public boolean redo(Term[] insts, Prolog p, ChoicePoint cp) throws PrologException {
	p.set_pending_goal(next);

	return unifier.redo(g.instantiateHead(insts), p, cp);
    }

    void instantiate(Term[] insts) {
	g = (Compound)g.instantiate(insts);

	if(next != null) {
	    next.instantiate(insts);
	}
    }

    Goal normalise(Variable[] vars,
 		   int next_new_var,
		   Vector new_vars,
		   TypeDictionary types) {
	// Copy the current goal + conjunction node.
	Term[] new_args   = new Term[g.arity];
	Compound new_g    = new Compound(g.functor, new_args);
	Conjunction new_c = new Conjunction(new_g, unifier);

	new_c.next = next;

	/* The body consists of unify()'s and calls to other
	 * predicates. Ensure the following conditions hold:
	 *
	 * 1. All unifications have the form: unify(X, S).
	 *      where: X is a variable
	 *             S is one of: variable
	 *                          number
	 *                          structure containing new variables.
	 * 2. Calls contain variables only.
	 */
	if(g.isUnify()) {
// 	    System.err.println("normalise: unification: " + g);

	    /* Ensure arg0 is a variable. */
	    if(g.args[1] instanceof Variable) {
		new_args[0] = g.args[1];
		new_args[1] = g.args[0];
	    } else {
		new_args[0] = g.args[0];
		new_args[1] = g.args[1];
	    }

	    if(!(new_args[0] instanceof Variable)) {
		throw new RuntimeException("normalise: unify/2 has nonvar arg0: " + new_g);
	    }

	    // Map arg0 to a "new" variable, if necessary.
	    int arg0_id = ((Variable)new_args[0]).id;

	    if(arg0_id < vars.length) {
		if(vars[arg0_id] == null) {
		    new_vars.addElement(new_args[0]
					= vars[arg0_id]
					= new Variable(next_new_var++));
		} else {
		    new_args[0] = vars[arg0_id];
		}
	    }

	    if(new_args[1] instanceof Compound) {
		Compound arg1        = (Compound)new_args[1];
		Term[] arg1_new_args = new Term[arg1.arity];
		new_args[1]          = new Compound(arg1.functor, arg1_new_args);

		Conjunction ptr = new_c;

//  		System.err.println("normalise: flatten compound term: " + arg1);

		for(int i = 0; i < arg1.arity; i++) {
		    Term arg = arg1.args[i];

		    Variable var = new Variable(next_new_var++);
		    arg1_new_args[i] = var;
		    new_vars.addElement(var);

		    if(arg instanceof Variable) {
			Variable v = (Variable)arg;

			if(vars[v.id] == null) {
			    vars[v.id] = var;
			    continue;
			}
		    }

// 		    System.err.println("  adding unify(" + var + ", " + arg + ")");

		    Conjunction unify = new Conjunction(new Compound(unify_2.functor,
								     new Term[]{var, arg}),
							unify_2.unify);

		    unify.next = ptr.next;
		    ptr.next = unify;
		    ptr = unify;
		}
	    }
	} else if(g.isIs()) {
	    // Arithmetic: just copy it.
	    Term[] g_args = g.args;

	    for(int i = 0; i < g_args.length; i++) {
		int old_new_vars_size = new_vars.size();

		new_args[i] = g_args[i].normalise_copy(vars, next_new_var, new_vars);

		next_new_var += new_vars.size() - old_new_vars_size;
	    }
	} else if(g.arity > 0) {
//  	    System.err.println("normalise: call: " + g);

 	    Conjunction ptr = new_c;

 	    for(int i = 0; i < g.arity; i++) {
// 		System.err.println("  considering arg[" + i + "] -> " + g.args[i]);

 		Term arg = g.args[i];

 		if(arg instanceof Variable) {
		    Variable var = (Variable)arg;

		    if(var.id < vars.length) {
			if(vars[var.id] == null) {
			    new_vars.addElement(new_args[i]
						= vars[var.id]
						= new Variable(next_new_var++));
// 			    System.err.println("  create new var -> " + new_args[i]);
			} else {
// 			    System.err.println("  update old var -> " + new_args[i]);
			    new_args[i] = vars[var.id];
			}
		    } else {
// 			System.err.println("  using new var -> " + new_args[i]);
			new_args[i] = arg;
// 			throw new RuntimeException("Arg: " + i + " -> " + arg);
		    }
 		} else {
 		    Variable var = new Variable(next_new_var++);
 		    new_args[i] = var;
		    new_vars.addElement(var);

// 		    System.err.println("  non-var arg -> " + var);

 		    /* Insert node before the call, i.e. move the call to a
                     * new node, recycle this one with the arg
                     * unification. */
 		    Compound unify_arg = new Compound(unify_2.functor,
 						      new Term[]{var, arg});

 		    Conjunction call = new Conjunction(ptr.g,
						       ptr.unifier);

		    call.next = ptr.next;

 		    ptr.g       = unify_arg;
		    ptr.unifier = unify_2.unify;
 		    ptr.next    = call;

 		    ptr = call;
 		}
 	    }

 	    /* If we didn't modify the call, proceed. */
 	    if(ptr != new_c) {
// 		typify(types, new_args);
// 		System.err.println(" -- after normalising the call:\n\t" + this);
		return new_c.normalise(vars, next_new_var, new_vars, types);
	    }
 	}

 	typify(types, new_args);

	if(next != null) {
	    new_c.next = new_c.next.normalise(vars, next_new_var, new_vars, types);
	}

	return new_c;
    }

    private void typify(TypeDictionary types, Term[] args) {
	for(int i = 0; i < args.length; i++) {
// 	    System.err.println(">> TYPIFY: pre: args[" + i + "] " + args[i]);
	    args[i].typify(types, false);
// 	    System.err.println(">> TYPIFY: " + args[i] + " -> " + args[i].ti);
	}
    }

    public boolean compile(ClauseCodeGen ccg) {
	boolean c = unifier.compile_call(ccg, g.args, next == null);

// 	System.err.println(ccg);

	return c && (next == null) ? true : next.compile(ccg);
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append(g.toString());

// 	if(unifier != null) {
// 	    out.append(" (unifier)");
// 	}

	if(next != null) {
	    out.append(",\n\t");
	    out.append(next.toString());
	}

	return out.toString();
    }
}
