package prolog;

import java.io.*;

import prolog.util.*;

/**
 * The Prolog database.
 * <P>
 * Hash of the functors of the heads + "/" + the arity of the head.
 * <P>
 * The database is scanned for a user definition of a predicate with the
 * given functor and arity. If this fails, an attempt is made to load an
 * external class with the same characteristics.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Database.java,v 1.24 2000/11/07 12:04:55 peteg Exp $
 */
public class Database implements Cloneable, Serializable {
    /**
     * The database itself is a hashtable. I've inlined the code as it's
     * easier to make a pile of assumptions about the implementation for
     * performance reasons.
     */
    private transient Predicate[] table;

    /**
     * The total number of entries in the hash table.
     */
    private transient int count;

    /**
     * Rehashes the table when count exceeds this threshold.
     */
    private int threshold;

    /**
     * The load factor for the hashtable.
     */
    private float loadFactor;

    /**
     * Are we just-in-time compiling clauses?
     */
    public boolean jit;

    /**
     * The queue of predicates to compile. Contains
     * <TT>PredicateCodeGen</TT> objects.
     */
    public Queue compilerQueue;

    /**
     * Should the compiler run in a separate thread, or should we wait for
     * it to compile something?
     */
//     public static final boolean asyncCompiler = true;
    public static final boolean asyncCompiler = false;

    /**
     * Creates an empty database.
     */
    public Database(boolean jit) {
	this(jit, 101);
    }

    /**
     * Creates a database with given initial size.
     */
    public Database(boolean jit, int dbsize) {
	if(dbsize <= 0) {
	    dbsize = 101;
	}

	table = new Predicate[dbsize];
	loadFactor = 0.75f;
	threshold = (int)(dbsize * loadFactor);
	count = 0;

	this.jit = jit;
	initCompiler();
    }

    private final void initCompiler() {
	JITThread jitt;

	if(asyncCompiler) {
	    jitt = new JITThread(compilerQueue = new ASyncQueue());
	    Thread thread = new Thread(jitt);
	    thread.setDaemon(true);
	    thread.start();
	} else {
	    jitt = new JITThread();
	    compilerQueue = new SyncQueue(jitt);
	}
    }

    /**
     * Finds a potential unifer for a given Term. 
     */
    public Predicate get(Compound c) {
	return get(c.functor, c.arity);
    }

    /** 
     * Finds a <TT>Predicate<TT> record with the given functor and arity. If
     * there isn't one in the database, attempt to load a pre-compiled one
     * from <TT>prolog.lib</TT>.
     *
     * @param f functor
     * @param a arity
     */
    public Predicate get(String f, int a) {
	int hash = f.hashCode() + a,
	    index = (hash & 0x7FFFFFFF) % table.length;
	Predicate p;

	for(p = table[index]; p != null; p = p.next) {
	    if(p.hash == hash && p.arity == a && p.functor == f) {
		return p;
	    }
	}

	/* Attempt to load an externally-compiled version. This is our
	 * standard library, though the semantics are wrong (we should
	 * ensure that predicates in the library cannot be redefined by the
	 * user. It is somewhat convenient for benchmarking that this isn't
	 * the case).
	 */
	if((p = load(f, a)) == null) {
	    return null;
	}

	return p;
    }

    /**
     * Loads an external class file which should subclass
     * <TT>prolog.CompiledPredicate</TT> and have a constructor with no
     * parameters.
     *
     * @see prolog.CompiledPredicate
     */
    Predicate load(String f, int a) {
	String name = "prolog.lib." + f + "_" + a;
	Predicate p;

	try {
	    p = (Predicate)Class.forName(name).newInstance();
	} catch(Exception e) {
 	    System.err.println("Error loading class: " + name + " : " + e);
	    return null;
	}

	put(f, a, p);
	System.err.println("Loaded class: " + name);

	return p;
    }

    /**
     * Rehashes the contents of the hashtable into a hashtable with a 
     * larger capacity. This method is called automatically when the 
     * number of keys in the hashtable exceeds this hashtable's capacity 
     * and load factor. 
     */
    private void rehash() {
	int oldCapacity = table.length;
	Predicate oldTable[] = table;

	int newCapacity = oldCapacity * 2 + 1;
	Predicate newTable[] = new Predicate[newCapacity];

	threshold = (int)(newCapacity * loadFactor);
	table = newTable;

	System.err.println("Database.rehash(): old = " + oldCapacity + ", new = " + newCapacity + ", thresh = " + threshold + ", count = " + count);

	for(int i = oldCapacity; i-- > 0; ) {
	    for(Predicate old = oldTable[i]; old != null; ) {
		Predicate e = old;
		old = old.next;

		int index = (e.hash & 0x7FFFFFFF) % newCapacity;
		e.next = newTable[index];
		newTable[index] = e;
	    }
	}
    }

    /**
     * Inserts a predicate into the database.
     */
    public void put(String f, int a, Predicate p) {
	int hash = f.hashCode() + a,
	    index = (hash & 0x7FFFFFFF) % table.length;

	Predicate[] tab = table;

// 	for(Predicate cur = tab[index]; cur != null; cur = cur.next)
// 	    if(cur.hash == hash && cur.arity == a && cur.functor.equals(f))
// 		throw new PrologException("Database.put(): predicate multiply defined: " + p);

	if(count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();
	    put(f, a, p);
	    return;
	} 

	// Creates the new entry.
	p.hash = hash;
	p.functor = f;
	p.arity = a;
	p.next = tab[index];

	tab[index] = p;
	count++;
    }

    /**
     * Inserts a a clause into the database. This interface should be used
     * to set up the database (when it is immaterial whether a predicate is
     * dynamic or not) and not for <TT>assert/1</TT>, <TT>retract/1</TT>,
     * etc.
     *
     * @param c the clause to be inserted.
     * @see prolog.lib.asserta_1
     * @see prolog.lib.assertz_1
     * @see prolog.lib.abolish_1
     * @see prolog.lib.retract_1
     */
    public void put(Clause c) {
	Compound head = c.head;
	String f = head.functor;
	int a = head.arity;
	InterpretedPredicate p = (InterpretedPredicate)get(f, a);

	if(p == null) {
	    p = newPredicate(a);
	    put(f, a, p);
	}

	p.putz(c);
    }

    /**
     * Clones this database. This is a deep copy. Note the implementation
     * uses the serialize/deserialize approach (i.e. it's inefficient) as
     * cloning the graph directly is tricky and efficiency isn't really a
     * priority for this op.
     * <P>
     * Reference: <A HREF="http://www.javaworld.com/javaworld/javatips/jw-javatip76.html">Java Tip 76: An alternative to the deep copy technique</A>.
     */
    public Object clone() {
 	try {
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    ObjectOutputStream oos = new ObjectOutputStream(bos);

	    oos.writeObject(this);
	    oos.flush();

	    return new ObjectInputStream(new ByteArrayInputStream(bos.toByteArray())).readObject();
	} catch(IOException e) {
	    throw new RuntimeException("Database.clone(): " + e);
	} catch(ClassNotFoundException e) {
	    throw new RuntimeException("Database.clone(): " + e);
	}
    }

    /**
     * writeObject is called to save the state of the hashtable to a stream.
     */
    private synchronized void writeObject(java.io.ObjectOutputStream s)
        throws IOException {
	// Write out the fields of primitive type.
	s.defaultWriteObject();

	// Write out length, count of elements, Predicates.
	s.writeInt(table.length);
	s.writeInt(count);
	for(int i = table.length; i-- > 0; )
	    for(Predicate p = table[i]; p != null; p = p.next)
		s.writeObject(p);
    }

    /**
     * readObject is called to restore the state of the hashtable from
     * a stream.
     */
    private synchronized void readObject(java.io.ObjectInputStream s)
	throws IOException, ClassNotFoundException {
	// Read in the length, threshold, and loadfactor
	s.defaultReadObject();

	// Read the original length of the array and number of elements
	int length = s.readInt();
	int elements = s.readInt();

	table = new Predicate[length];
	count = 0;

	// Read the number of elements and then all the key/value objects
	while(elements-- > 0) {
	    Predicate p = (Predicate)s.readObject();
	    put(p.functor, p.arity, p);
	}
    }

    /**
     * Lists all clauses in the database.
     */
    public String toString() {
 	StringBuffer out = new StringBuffer();

	for(int i = table.length; i-- > 0; ) {
	    for(Predicate cur = table[i]; cur != null; cur = cur.next) {
		out.append("% ");
		out.append(cur.functor);
		out.append("/");
		out.append(cur.arity);
		out.append(cur.toString());
		out.append("\n");
	    }
	}

// 	out.append(JITCompiler.classloader);

 	return out.toString();
    }

    /**
     * Create a new <TT>InterpretedPredicate</TT> record.
     */
    InterpretedPredicate newPredicate(int arity) {
// 	return new InterpretedPredicate(arity);
 	return jit ? new JITPredicate(arity) : new InterpretedPredicate(arity);
    }
}
