package prolog;

/* The gnu.getopt package. */
import gnu.getopt.LongOpt;
import gnu.getopt.Getopt;

import java.io.*;

/**
 * The application top-level. Parse the command tail, initialise a prolog
 * engine, support the interactive session.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Main.java,v 1.10 2000/08/20 15:31:07 peteg Exp $
 */
public class Main {
    /** 
     * Reads from the filename specified on the command line, executes the
     * predicate <TT>main/0</TT> if defined, otherwise enters a read-eval
     * loop. Note that I use <A HREF="http://www.urbanophile.com/arenn/hacking/download.html">gnu.getopt</A>
     * to parse command line arguments.
     */
    public static void main(String argv[]) {
	try {
	    Prolog pt = parseCommandTail(argv, null);

	    /* Fall into the read-eval loop, maybe because the file read
	     * failed, or there is no main/0 predicate.
	     */
	    readeval(pt);
	    System.err.println();
	} catch(Exception e) {
// 	    System.err.println(e);
	    e.printStackTrace();
	}
    }

    /**
     * Command-line argument parser.
     *
     * @param o an application-specific object. Gives custom
     * (Java-implemented) predicates access to some global state.
     * @exception prolog.PrologException Generated by the interpreter on
     * e.g. type exceptions and the parser on syntax errors.
     */
    public static Prolog parseCommandTail(String[] argv, Object o)
	throws PrologException {

	Prolog pt;

	boolean jit = false;
	int dbsize = 0;
	boolean time = false;
	String database = null;

	/* Print an obnoxious header. */
	System.err.println("Prolog 0.005\n" +
			   "(C)opyright 2000 Peter Gammie.\n" +
			   "Prolog is free software, covered by the GNU Library General Public License, and\n" +
			   "you are welcome to change it and/or distribute copies of it under certain\n" +
			   "conditions.");

	/* Parse the command tail. */
	LongOpt[] longopts = {
	    new LongOpt("benchmark", LongOpt.NO_ARGUMENT, null, 'b'),
	    new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h'),
	    new LongOpt("initdbsize", LongOpt.REQUIRED_ARGUMENT, null, 'i'),
	    new LongOpt("jit", LongOpt.NO_ARGUMENT, null, 'j'),
	    new LongOpt("loaddb", LongOpt.REQUIRED_ARGUMENT, null, 'l'),
	};

	Getopt g = new Getopt("prolog", argv, "bhi:jl:", longopts);
	int c;
	String arg;
	g.setOpterr(false);

	while((c = g.getopt()) != -1) {
	    switch(c) {
	    case 'b':
		System.err.println("Benchmark option set.");
		time = true;
		break;
	    case 'h':
		System.err.println("\nUsage: prolog.Main [OPTION]... [FILE]...\n\n" +
				   "  -b, --benchmark\tTime each top-level query.\n" +
				   "  -h, --help\t\tDisplay this help and exit.\n" +
				   "  -i, --initdbsize SIZE\tThe initial database size (in terms of predicates).\n" +
				   "  -j, --jit\t\tEnable the just-in-time prolog compiler.\n" +
				   "  -l, --loaddb FILE\tLoad a serialized database.\n");
		    System.exit(1);
	    case 'i':
		arg = g.getOptarg();
		try {
		    dbsize = Integer.parseInt(arg);
		} catch (NumberFormatException e) {
		    System.err.println("prolog: invalid parameter to initdbsize: " + arg);
		    System.err.println("Try `prolog.Main --help' for more information.");
		    System.exit(1);
		}
		break;
	    case 'j':
		System.err.println("JIT option set.");
		jit = true;
		break;
	    case 'l':
		database = g.getOptarg();
		break;
	    case '?':
		System.err.println("prolog: invalid option " + (char)g.getOptopt());
		System.err.println("Try `prolog.Main --help' for more information.");
		System.exit(1);
	    default:
		System.err.print("getopt() returned " + (char)c + "\n");
		System.exit(1);
	    }
	}

	/* Create the prolog thread given the database specs. */
	if(database != null) {
	    Database db;

	    /* Deserialize the database. */
	    System.err.println("Loading serialized database: " + database);
	    try {
		/* Standard technique - straight out of "Java in a Nutshell" */
		FileInputStream fis = new FileInputStream(database);
		ObjectInputStream in = new ObjectInputStream(fis);
		db = (Database)in.readObject();
		pt = new Prolog(db);
	    } catch (ClassNotFoundException e) {
		System.err.print("prolog.Main: class not found.");
		e.printStackTrace();
		throw new PrologException("prolog.Main: ClassNotFoundException.");
	    } catch (IOException e) {
		System.err.println("prolog.Main: error accessing file.");
		e.printStackTrace();
		throw new PrologException("prolog.Main: IOException.");
	    } 
	} else if(dbsize > 0) {
	    /* Initial database size. */
	    System.err.println("Initial database size: " + dbsize + " entries.");
	    pt = new Prolog(jit, dbsize);
	} else {
	    pt = new Prolog(jit);
	}

	pt.time = time;

	/* Stash the object if needs be. */
	pt.obj = o;

	/* Read all filenames specified -- they're "non-option"s. */
	try {
	    for(int i = g.getOptind(); i < argv.length; i++) {
		System.err.print("Loading: " + argv[i]);
		pt.consult(argv[i]);
		System.err.println(" (done)");
	    }

	    /* Execute main/0 if defined. Look for a single solution. */
	    if(pt.db.get("main", 0) != null) {
		pt.once(pt.toGoal(new Compound("main")));
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}

	return pt;
    }

    /**
     * Read-eval loop. Query user for commands, execute them.
     */
    public static void readeval(Prolog pt) {
	while(true) {
	    System.err.print(": ");

	    try {
		if(pt.parser.parse(pt, Prolog.sys_lex, true)) {
		    // EOF
		    break;
		}
	    } catch(IOException e) {
		System.err.println(e);
		break;
	    } catch(PrologException e) {
		System.err.println(e);
		continue;
	    }
	}
    }
}
