package prolog;

import java.util.Enumeration;

import prolog.util.Hashtable;

/**
 * The table of operator precedences.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: OperatorTable.java,v 1.12 2000/08/17 13:01:23 peteg Exp $
 * @see prolog.Operator
 */
public class OperatorTable {
    /**
     * The standard operators (maps <TT>String -> Operator</TT>).
     */
    private static final OperatorTable stdops;

    static {
	stdops = new OperatorTable(new Hashtable());

	try {
	    stdops.put(1200, Operator.XFX, ":-");
	    stdops.put(1200, Operator.XFX, "-->");

	    stdops.put(1200, Operator.FX , ":-");

	    stdops.put(1100, Operator.XFY, ";");

	    stdops.put(1050, Operator.XFY, "->");

	    stdops.put(1000, Operator.XFY, ",");

	    stdops.put( 900,  Operator.FY , "\\+", 	"not");

	    stdops.put( 700, Operator.XFX, "=",		"unify");
	    stdops.put( 700, Operator.XFX, "\\=",	"not_unify");

	    stdops.put( 700, Operator.XFX, "==",	"identical");
	    stdops.put( 700, Operator.XFX, "\\==",	"not_identical");
	    stdops.put( 700, Operator.XFX, "@<");
	    stdops.put( 700, Operator.XFX, "@>");
	    stdops.put( 700, Operator.XFX, "@=<");
	    stdops.put( 700, Operator.XFX, "@>=");

	    stdops.put( 700, Operator.XFX, "=..",	"univ");

	    stdops.put( 700, Operator.XFX, "is",	"is");
	    stdops.put( 700, Operator.XFX, "=:=",	"is_equal");
	    stdops.put( 700, Operator.XFX, "=\\=",	"is_not_equal");
	    stdops.put( 700, Operator.XFX, "<",		"lt");
	    stdops.put( 700, Operator.XFX, "=<",	"le");
	    stdops.put( 700, Operator.XFX, ">",		"gt");
	    stdops.put( 700, Operator.XFX, ">=", 	"ge");

	    stdops.put( 500, Operator.YFX, "+");
	    stdops.put( 500, Operator.YFX, "-");
	    stdops.put( 500, Operator.YFX, "/\\");
	    stdops.put( 500, Operator.YFX, "\\/");

	    stdops.put( 400, Operator.YFX, "*");
	    stdops.put( 400, Operator.YFX, "/");
	    stdops.put( 400, Operator.YFX, "//");
	    stdops.put( 400, Operator.YFX, "rem");
	    stdops.put( 400, Operator.YFX, "mod",	"%");
	    stdops.put( 400, Operator.YFX, "<<");
	    stdops.put( 400, Operator.YFX, ">>");

	    stdops.put( 200, Operator.XFY, "**");

	    stdops.put( 200, Operator.XFY, "^");

	    stdops.put( 200, Operator.FY , "+");
	    stdops.put( 200, Operator.FY , "-");

	    stdops.put( 200, Operator.FY , "\\");

	    /* Extra def's, not in the standard. */
	    stdops.put(1175, Operator.XFX, "::=",	"_type_def");
	    stdops.put(1175, Operator.XFX, "::",	"_has_type");

	} catch(ArgumentException e) {
	    throw new InternalError("this shouldn't happen");
	}
    }

    /**
     * The operator table. Maps <TT>String -> Operator</TT>.
     */
    private Hashtable ops;

    OperatorTable() {
	this.ops = (Hashtable)stdops.ops.clone();
    }

    /**
     * The bootstrap constructor, used to init <TT>stdops</TT>.
     */
    private OperatorTable(Hashtable ops_hash) {
	this.ops = ops_hash;
    }

    /**
     * Add an operator.
     *
     * @param lexeme the string representing this operator.
     * @param type one of the associativity specifiers (XFX, ...) as a String.
     * @param p the precedence of this operator.
     * @exception prolog.ArgumentException on invalid precedence.
     * @see prolog.Operator
     * @see prolog.lib.op_3
     */
    public void put(int p, String type, String lexeme)
	throws ArgumentException {

	int t;

	type = type.toLowerCase();

	if(type.equals(Operator.S_FX)) {
	    t = Operator.FX;
	} else if(type.equals(Operator.S_FY)) {
	    t = Operator.FY;
	} else if(type.equals(Operator.S_XF)) {
	    t = Operator.XF;
	} else if(type.equals(Operator.S_YF)) {
	    t = Operator.YF;
	} else if(type.equals(Operator.S_XFX)) {
	    t = Operator.XFX;
	} else if(type.equals(Operator.S_XFY)) {
	    t = Operator.XFY;
	} else if(type.equals(Operator.S_YFX)) {
	    t = Operator.YFX;
	} else {
	    throw new ArgumentException("OperatorTable.put(): unknown type: " + type);
	}

	put(p, t, lexeme);
    }

    /**
     * Add an operator. No translation performed - the <TT>name</TT> of this
     * operator is the same as it's lexical representation.
     *
     * @param lexeme the string representing this operator.
     * @param type one of the associativity specifiers (XFX, ...).
     * @param p the precedence of this operator.
     * @exception prolog.ArgumentException on invalid precedence.
     * @see prolog.Operator
     */
    void put(int p, int type, String lexeme) throws ArgumentException {
	put(p, type, lexeme, lexeme);
    }

    /**
     * Add an operator - the functor of the compound term will be
     * <TT>name</TT>, rather than <TT>lexeme</TT>.
     *
     * @param lexeme the string representing this operator.
     * @param name the functor of the compound built to describe this term.
     * @param type one of the associativity specifiers (XFX, ...).
     * @param p the precedence of this operator.
     * @exception prolog.ArgumentException on invalid precedence.
     * @see putprolog.Operator
     */
    void put(int p, int type, String lexeme, String name)
	throws ArgumentException {

	if(p == Operator.min_precedence) {
	    // Remove operator. FIXME
	    throw new RuntimeException("put(): remove operator. not implemented.");
	} else if(p <= Operator.max_precedence) {
	    Operator op;

	    switch(type) {
	    case Operator.FX:
		op = new Operator(lexeme, name, Operator.prefix,  p, 0  , p-1);
		break;
	    case Operator.FY:
		op = new Operator(lexeme, name, Operator.prefix,  p, 0  , p);
		break;
	    case Operator.XF:
		op = new Operator(lexeme, name, Operator.postfix, p, p-1, 0);
		break;
	    case Operator.YF:
		op = new Operator(lexeme, name, Operator.postfix, p, p  , 0);
		break;
	    case Operator.XFX:
		op = new Operator(lexeme, name, Operator.infix,   p, p-1, p-1);
		break;
	    case Operator.XFY:
		op = new Operator(lexeme, name, Operator.infix,   p, p-1, p);
		break;
	    case Operator.YFX:
		op = new Operator(lexeme, name, Operator.infix,   p, p,   p-1);
		break;
	    default:
		throw new ArgumentException("OperatorTable.add(" + lexeme + ", " + name + "): unknown type: " + type);
	    }

	    /* Chain to existing operator. */
	    op.next = (Operator)ops.get(lexeme);
	    ops.put(lexeme, op);
	} else {
	    throw new ArgumentException("invalid precedence: " + p);
	}
    }

    /**
     * Finds an operator with this symbol and that type (XFX, ...).
     *
     * @param type one of the associativity specifiers (XFX, ...).
     * @see prolog.Operator
     */
    Operator lookup(String lexeme, int type) {
	for(Operator op = (Operator)ops.get(lexeme); op != null; op = op.next) {
	    if(op.type == type) {
		return op;
	    }
	}

	return null;
    }

    Operator lookup(String lexeme) {
	return (Operator)ops.get(lexeme);
    }

    /**
     * Remove an operator.
     */
    void delete(String lexeme, int type) {
	for(Operator op = (Operator)ops.get(lexeme), prev = null;
	    op != null;
	    prev = op, op = op.next) {

	    if(op.type == type) {
		if(prev == null) {
		    if(op.next == null) {
			ops.remove(lexeme);
		    } else {
			ops.put(lexeme, op.next);
		    }
		} else {
		    prev.next = op.next;
		}

		break;
	    }
	}
    }

    /**
     * Lists all operators in this table.
     */
    public String toString() {
 	StringBuffer out = new StringBuffer();

	out.append("--- OperatorTable: operators ---\n");
	for(Enumeration e = ops.elements(); e.hasMoreElements(); ) {
	    out.append(e.nextElement());
	    out.append("\n");
	}

	return out.toString();
    }
}
