package prolog;

import java.io.*;
import java.util.Enumeration;

import prolog.math.*;
import prolog.util.Hashtable;
import prolog.util.Vector;
import prolog.util.Strings;

/**
 * An operator-shift-precedence parser for ISO Prolog. Based loosely on that
 * of <A HREF="http://www.gnu.org/software/prolog/">GNU Prolog</A> by Daniel
 * Diaz.
 * <P>
 * Note that the externally-callable methods are not re-entrant, as they
 * rely on fields to implement synthetic attributes.
 * 
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Parser.java,v 1.25 2000/11/12 12:22:45 peteg Exp $
 * @see prolog.ParsingException
 */
public final class Parser {
    public static final String QUERY_FUNCTOR = ":-".intern();
    public static final int QUERY_ARITY   = 1;

    public static final String NECK_FUNCTOR  = ":-".intern();
    public static final int NECK_ARITY    = 2;

    /**
     * The Runtime object for this application. Used to get stats, etc.
     */
    private static Runtime runtime = Runtime.getRuntime();

    /**
     * The end-of-list functor.
     */
    public static final String empty_list_functor = "[]".intern();

    public static final String list_functor = ".".intern();

    /**
     * The operator table for this parser.
     */
    public final OperatorTable ops = new OperatorTable();

    /**
     * Indicates that there are no variables in this <TT>Compound</TT>
     * structure.
     */
    private boolean ground;

    /**
     * The set of variables in the current clause.
     */
    private Hashtable vardb = new Hashtable();

    /**
     * The lexer.
     */
    private Yylex lex;

    /* Contexts */
    private static int
	GENERAL_TERM  = 0,
	TRYING_PREFIX = 1,
	INSIDE_ANY_OP = 2;

    public Parser() {
    }

    /**
     * Parses some prolog text and interprets it in the context of the given
     * <TT>Prolog</TT> engine. This may involve asserting a clause,
     * answering a query, etc. Note that only one clause/query/etc. is dealt
     * with at a time.
     *
     * @param interactive print out variable bindings on success.
     * @return <TT>true</TT> on end-of-file, <TT>false</TT> otherwise.
     */
    public boolean parse(Prolog p, Yylex lex, boolean interactive)
	throws IOException, PrologException {

// 	System.err.println("Parser.parse()");

	Term t;

	if((t = parse(lex)) == null) {
	    return lex.eof;
	}

	if(t instanceof Compound) {
	    Compound c = (Compound)t;
	    Goal goal = null;

	    if(interactive) {
		goal = p.toGoal(c);
	    }

	    if(c.functor == QUERY_FUNCTOR && c.arity == QUERY_ARITY) {
		goal = p.toGoal((Compound)c.args[0]);
	    }

	    if(goal != null) {
		// Query
		query(p, goal, interactive);
	    } else {
		// Assert
		Compound head;
		Goal body;

		if(c.functor == NECK_FUNCTOR && c.arity == NECK_ARITY) {
		    head = (Compound)c.args[0];
		    body = p.toGoal((Compound)c.args[1]);
		} else {
		    head = c;
		    body = null;
		}

// 		System.err.println("Assert: " + c + "\n" + clause);

		p.db.put(new Clause(head, body,
				    c,
				    vardb.size()));
	    }

	    return lex.eof;
	}

	throw new PrologException("main(): dunno what to do with: " + t);
    }

    /**
     * Finds the solutions to a query.
     */
    private void query(Prolog p, Goal goal, boolean interactive)
	throws PrologException {

	long startTime, queryTime, gcTime;
	boolean succeeded;

	Term[] insts = new Term[vardb.size()];
	goal.instantiate(insts);

// 	do { 
	    System.err.println("Query: " + goal);
// 	    System.err.println("Prolog.find(" + goal + ", " + all + ")");

	    startTime = System.currentTimeMillis();
	    succeeded = p.find(goal);
	    queryTime = System.currentTimeMillis();
	    runtime.gc();
	    gcTime    = System.currentTimeMillis();

// 	    System.err.println(p);

	    if(interactive) {
		if(succeeded) {
		    /* Successful resolution, print results. */
		    System.err.println("** yes");

		    for(int i = 0; i < insts.length; i++) {
 			System.err.println("  _" + i + " = "
					   + insts[i].toString());
		    }
		} else {
		    System.err.println("** no");
		}
	    }

	    /* ...and statistics if asked for. */
 	    if(p.time) {
		gcTime    -= queryTime;
		queryTime -= startTime;

		long totalmem = runtime.totalMemory();
		long currentmem = runtime.freeMemory();
		long percent = (100 * (totalmem - currentmem)) / totalmem;
		totalmem = (long) (((float) totalmem) / 1000000.0);

		System.err.println("(query: " + queryTime + "ms, gc: " + gcTime +
				   "ms, memory used: " +
				   percent + "% out of " + totalmem + " Mb)");
 	    }

	    /* Find the next solution. */
// 	    goal = null;
// 	} while(succeeded);
    }

    /**
     * Parses a term without interpreting it.
     *
     * @param lex the lexical analyzer from which we read tokens.
     * @exception IOException on stream errors.
     * @exception prolog.ParsingException on syntax/semantical errors.
     */
    public Term parse(Yylex lex) throws IOException, ParsingException {
// 	System.err.println("Parser.parse(): lex.eof: " + lex.eof);

	if(lex.eof) {
	    return null;
	}

	vardb.clear();
	this.lex = lex;
	Term term;

	if((term = term(Operator.max_precedence, GENERAL_TERM, false)) == null) {
	    // If it's eof, return null, otherwise spit the dummy.
	    if(lex.eof) {
		return null;
	    } else {
		throw new ParsingException("Parser.parse(): got null term but not EOF!");
	    }
	}

// 	System.err.println("Parser.parse(): lex: " + lex);

	lex.scan(false);

	if(lex.tokenType != Yylex.FULL_STOP) {
	    throw new ParsingException(lex, ". or operator expected after expression");
	}

	return term;
    }

    private Term term(int cur_prec, int context, boolean comma_is_punct)
	throws IOException, ParsingException {

	boolean left_is_op = false;
	int cur_left = 0;
	Operator op;
	String functor = null;
	Term term;

	lex.scan(context != TRYING_PREFIX || comma_is_punct);

// 	System.err.println("Parser.term("
// 			   + cur_prec + ", "
// 			   + context + ", "
// 			   + comma_is_punct + "): lex: " + lex);

	switch(lex.tokenType) {
	case Yylex.AVARIABLE:
//   	    System.err.println("Parser.term: got AVARIABLE");

	    ground = false;

	    term = new Variable(vardb.size());
	    vardb.put("_" + (vardb.size()), term);

	    break;

	case Yylex.VARIABLE:
//   	    System.err.println("Parser.term: got VARIABLE");

	    ground = false;

	    /* Reuse old variable definitions */
 	    if((term = (Variable)vardb.get(lex.stringVal)) == null) {
 		vardb.put(lex.stringVal, term = new Variable(vardb.size()));
 	    }

	    break;

	case Yylex.DOUBLE:
//     	    System.err.println("Parser.term: got DOUBLE");
 	    term = new FloNum(lex.doubleVal);
	    break;

	case Yylex.INTEGER:
//   	    System.err.println("Parser.term: got INTEGER");
	    term = IntNum.make(lex.intVal);
	    break;

	case Yylex.PUNCTUATION:
	    if("({[".indexOf(lex.charVal) == -1) {
		lex.unget();
		return null;
// 		throw new ParsingException(lex, "punctuation but not a scoping symbol.");
            }

//  	    atom = (lex.charVal == '{') ? atom_curly_brackets : ATOM_NIL;

	    if((term = parse_Bracketed_Term()) == null) {
// 		System.err.println("  Parser.term: [] returned.");
// 		term = EmptyList.emptyList;
		term = newCompound(empty_list_functor, (Term[])null);
	    }

	    break;

	case Yylex.CUT:
	    /* Special case the cut. Probably not ISO-compliant. */
	    term = prolog.lib._cut_0.cut;
	    break;

	case Yylex.STRING:
// 	    System.err.println("Parser.term: got STRING: " + lex.stringVal);
	    functor = lex.stringVal;

	    lex.scan(false);

	    if(lex.tokenType == Yylex.PUNCTUATION && lex.charVal == '(') {
// 		System.err.println(" >> treating as a compound, looking for args.");
		term = newCompound(functor, args());
		break;
	    } else {
		lex.unget();
	    }
                                                /* maybe a prefix operator */
	    if((op = ops.lookup(functor, Operator.prefix)) != null
	       && (cur_prec >= op.precedence)) {
                                                  /* try a prefix operator */
		cur_left = op.precedence;
		if((term = term(op.right, TRYING_PREFIX, comma_is_punct)) != null) {
// 		    System.err.println(functor + " is a prefix op, it's name is: " + op.name);
		    term = newCompound(op.name, new Term[]{term});
		    break;
		}
                                              /* prefix operator as a name */
		if(context != GENERAL_TERM) {
		    throw new ParsingException(lex, "expression expected or previous operator needs brackets");
		}
	    }

	    left_is_op = ops.lookup(functor) != null;

	    if(left_is_op && context != GENERAL_TERM) {  /* in operator context */
		if(ops.lookup(functor, Operator.infix) != null
		   || ops.lookup(functor, Operator.postfix) != null) {
		    throw new ParsingException(lex, "current or previous operator needs brackets");
		} else {
		    throw new ParsingException(lex, "current operator needs brackets");
		}
	    }

	    term = newCompound(functor, (Term[])null);
	    break;

	case Yylex.EOF:
	    return null;

	default:   /* FULL_STOP */
	    throw new ParsingException(lex, "unknown token.");
	}

// 	System.err.println("Parser.term() -- halfway -- term = " + term);

	while(true) {
	    boolean infix_op;

	    lex.scan(comma_is_punct);

// 	    System.err.println("Parser.term2("
// 			       + cur_prec + ", "
// 			       + context + ", "
// 			       + comma_is_punct + "): lex: " + lex);

	    if(lex.tokenType != Yylex.STRING) {
		break;
	    }

	    functor = lex.stringVal;

	    if((op = ops.lookup(lex.stringVal, Operator.infix)) != null) {
// 		System.err.println("  " + lex.stringVal + " is an infix op.");
		infix_op = true;
	    } else if((op = ops.lookup(lex.stringVal, Operator.postfix)) != null) {
// 		System.err.println("  " + lex.stringVal + " is a postfix op.");
		infix_op = false;
	    } else {
		break;
	    }

	    if(left_is_op) {
		throw new ParsingException(lex, "previous operator needs brackets");
	    }

	    if(cur_prec < op.precedence || cur_left > op.left) {
		break;
	    }

	    if(infix_op) {	/* infix operator */
// 		System.err.println("Parser.term infix ("
// 				   + cur_prec + ", "
// 				   + context + ", "
// 				   + comma_is_punct + "): lex: " + lex);

		Term[] w = {
		    term,
		    term(op.right, INSIDE_ANY_OP, comma_is_punct)
		};

 		if(w[1] == null) {
 		    throw new ParsingException(lex, "right operand expected for infix operator");
 		}

 		term = newCompound(op.name, w);
	    } else {
 		term = newCompound(op.name, new Term[]{term}); /* postfix operator */
	    }

	    cur_left = op.precedence;
	}

	lex.unget();

	return term;
    }

    /**
     * Parse an argument list
     *
     * @exception prolog.ParsingException syntactic error
     */
    private Vector args() throws IOException, ParsingException {
	Vector args = new Vector();

	do {
	    args.addElement(term(Operator.max_arg_of_functor_prec,
				 GENERAL_TERM,
				 true));
	    lex.scan(true);
	} while(lex.tokenType == Yylex.PUNCTUATION && lex.charVal == ',');

	if(lex.tokenType != Yylex.PUNCTUATION || lex.charVal != ')') {
	    throw new ParsingException(lex, "closing parenthesis expected.");
	}

// 	System.err.println("Parser.args(): lex: " + lex);

	return args;
    }

    private Term parse_Bracketed_Term()	throws IOException, ParsingException {
	Term term;

	switch(lex.charVal) {
	case '(':
	    term = term(Operator.max_precedence, GENERAL_TERM, false);

	    lex.scan(false);

	    if(lex.tokenType != Yylex.PUNCTUATION || lex.charVal != ')') {
		throw new ParsingException(lex, ") or operator expected");
	    }

	    break;

//  	case '{':
//  	    term = term(lex, vardb,
//  			Operator.max_precedence, GENERAL_TERM, false);

//  	    lex.scan(false);

//  	    if(lex.tokenType != PUNCTUATION || lex.charVal != '}') {
//  		throw new ParsingException(lex, "} or operator expected");
//  	    }

// 	    if(term != NOT_A_WAM_WORD)	/* term==NOT_A_WAM_WORD if {} */
// // 		term = Mk_Compound(atom_curly_brackets, 1, &term);
// 	    break;

	case '[':
	    term = parse_List(true);	/* term==NOT_A_WAM_WORD if [] */
	    break;

	default:
	    throw new ParsingException(lex, "parse_Bracketed_Term - default.");
	}

	return term;
    }

    /**
     * Parses prolog's list syntax.
     *
     * @param can_be_empty the list can be empty, i.e. the <TT>[]</TT>.
     */
    private Term parse_List(boolean can_be_empty) throws IOException, ParsingException {
	Term car, cdr;

// 	System.err.println("parse_List() -- start. can_be_empty: " + can_be_empty);

	car = term(Operator.max_arg_of_functor_prec, GENERAL_TERM, true);

// 	System.err.println("parse_List(): car: " + car + " lex: " + lex);

	lex.scan(true);

// 	System.err.println("parse_List() -- 2 -- lex: " + lex);

	if(car == null) {
	    if(!can_be_empty) {
		throw new ParsingException(lex, "expression expected in list");
	    } else if(lex.tokenType != Yylex.PUNCTUATION
		      || lex.charVal != ']') {
		throw new ParsingException(lex, "expression or ] expected in list");
	    }

	    return null;
	}

// 	System.err.println("parse_List() -- 3 -- lex: " + lex);

	if(lex.tokenType != Yylex.PUNCTUATION || ",|]".indexOf(lex.charVal) == -1) {
	    throw new ParsingException(lex, ", | ] or operator expected in list");
	}

// 	System.err.println("parse_List() -- halfway -- lex: " + lex);

	switch(lex.charVal) {
	case ',':	/* [X,[...]] */
	    cdr = parse_List(false);
	    break;

	case '|':	/* [X|Y] */
	    cdr = term(Operator.max_arg_of_functor_prec, GENERAL_TERM, false);

 	    if(cdr == null) {
 		throw new ParsingException(lex, "cdr expression expected in list");
 	    }

	    lex.scan(true);

	    if(lex.tokenType != Yylex.PUNCTUATION || lex.charVal != ']') {
		throw new ParsingException(lex, "] or operator expected in list");
	    }

	    break;

	case ']':	/* [X] */
// 	    cdr = EmptyList.emptyList;
	    cdr = newCompound(empty_list_functor, (Term[])null);
	    break;

	default:
	    throw new ParsingException("parse_List - default.");
	}

// 	System.err.println("parse_List(): [" + car + "|" + cdr + "] lex: " + lex);

// 	return new List(car, cdr);
	return newCompound(list_functor, new Term[]{car, cdr});
    }

    private final Compound newCompound(String f, Vector arglist) throws ParsingException {
	Term[] args = null;

	if(arglist != null) {
	    args = new Term[arglist.size()];
	    arglist.trimToSize();
	    arglist.copyInto(args);
	}

	return newCompound(f, args);
    }

    /**
     * Create new Compounds. Factory method so that we normalise strings as
     * need be.
     */
    private final Compound newCompound(String f, Term[] arglist) throws ParsingException {
	String functor = Strings.expand_escapes(f);
	int arity = arglist == null ? 0 : arglist.length;
	Compound c;

// 	if(functor == EmptyList.emptyListFunctor && arity == 0) {
// 	    c = EmptyList.emptyList;
// 	} else if(functor == List.listFunctor && arity == 2) {
// 	    c = new List(arglist);
// 	} else {
	    c = new Compound(functor, arglist);
// 	}

	c.copied = ground;

// 	System.err.println(" >> newCompound(): " + c);

	return c;
    }

    /**
     * Test method - parse <TT>stdin</TT> and print out the resultant tree.
     */
    public static void main(String[] argv) throws Exception {
	Yylex lex = new Yylex(System.in);
	Parser parser = new Parser();

	while(true) {
	    System.err.println(">> Parser.main(): " + parser.parse(lex));
	}
    }
}
