package prolog;

import java.io.*;

/**
 * The abstract base class of the predicate representations - essentially
 * just a bucket in a hashtable.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Predicate.java,v 1.19 2000/11/07 12:04:56 peteg Exp $
 * @see prolog.ClauseIndex
 * @see prolog.CompiledPredicate
 * @see prolog.DynamicPredicate
 * @see prolog.InterpretedPredicate
 * @see prolog.JITPredicate
 */
public abstract class Predicate implements Cloneable, Serializable {
    /* Hashtable stuff. */
    transient int hash;
    public String functor;
    public int arity;

    transient Predicate next;

    protected Predicate() {
    }

    /**
     * Resolve the given goal in the context of a Prolog engine.
     *
     * @param goal the goal to be resolved.
     */
    public abstract boolean unify(Term[] args, Prolog p)
	throws PrologException;

    /**
     * Redo a goal, i.e. on backtracking.
     */
    public abstract boolean redo(Term[] args, Prolog p, ChoicePoint cp)
	throws PrologException;

    /**
     * Compiles a call to this predicate. This method should be overridden
     * by other classes that can do a better job than a generic call.
     *
     * @param lastcall this call is a last call.
     */
    public boolean compile_call(ClauseCodeGen ccg, Term[] args, boolean lastcall) {
//  	System.err.println("Predicate.compile_call()");

	return ccg.compile_call(this, args, lastcall);
    }

    /**
     * Object serialisation breaks the String.intern() method, which we rely
     * on to speed up functor equality checks (pointer equality v's string
     * equality). So we fix that up here...
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException {
	in.defaultReadObject();
	functor = functor.intern();
    }

    public abstract String toString();
}
