package prolog;

import java.io.*;

/**
 * The Prolog engine. This class implements the process of evaluating a goal
 * list.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Prolog.java,v 1.40 2000/11/12 12:22:45 peteg Exp $
 */
public final class Prolog {
    /**
     * Lexer for stdin, for interactive mode.
     */
    public final static Yylex sys_lex = new Yylex(System.in);

    /**
     * Contains all user- and system-defined types.
     */
    public final TypeDictionary types = new TypeDictionary();

    /**
     * The <TT>fail</TT> goal - used to force backtracking when looking for
     * all solutions.
     */
    private Goal fail;
//     private final Goal fail;

    /**
     * The current input stream.
     */
    public Yylex input_stream = sys_lex;

    /**
     * The trail records variable bindings that need to be undone on
     * backtracking.
     */
    private Binding trail;

    /**
     * The stack of <TT>ActivationRecord</TT>s.
     */
    private ActivationRecord[] runtime_stack;

    /**
     * Points to the top of the <TT>runtime_stack</TT>.
     */
    private int runtime_stack_top;

    private int ar_cur_index;
    private ActivationRecord ar_cur;
    private int old_cp_stack_top;
    private Goal goal, pending_goal;

    /**
     * Points to the last <TT>ActivationRecord</TT> on the
     * <TT>runtime_stack</TT> that has a <TT>ChoicePoint</TT> associated
     * with it.
     */
    private int cp_last;

    /**
     * The stack of <TT>ChoicePoint</TT>s.
     */
    private ChoicePoint[] cp_stack;

    /**
     * Points to the top of the <TT>cp_stack</TT>.
     */
    private int cp_stack_top;

    /**
     * The "cut pointer" B0 - points to the entry on the <TT>cp_stack</TT>
     * that was current when this predicate was invoked.
     */
    private int cut_ptr;

    /**
     * The parser for this <TT>Prolog</TT> engine.
     */
    public Parser parser;

    /**
     * The database of clauses.
     */
    public Database db;

    /**
     * An application-specific object (whatever it is).
     */
    public Object obj;

    /* Runtime options. */

    /**
     * Enable a crude benchmark - display the time taken to answer a query
     * and collect the resultant garbage. Highly dependent on system
     * load.
     */
    public boolean time = false;

    /**
     * The threshold at which we compile a predicate.
     */
    public final int threshold = 2;

    /**
     * Creates a new Prolog engine with default parameters.
     */
    public Prolog(boolean jit) {
	this(jit, -1);
    }

    /**
     * Creates a new Prolog engine with the specified initial database type
     * and size.
     */
    public Prolog(boolean jit, int dbsize) {
	this(new Database(jit, dbsize));
    }

    /**
     * Creates a new Prolog engine with a pre-initialised database
     * (e.g. recovered by deserialization).
     */
    public Prolog(Database dbin) {
	trail = null;

	runtime_stack = new ActivationRecord[100000];
	runtime_stack_top = 0;

	cp_last = -1;

	cp_stack = new ChoicePoint[1000];
	cp_stack_top = cut_ptr = 0;

	db = dbin;
	parser = new Parser();

	fail = new Conjunction(new Compound("fail"), null, this);
    }

    /**
     * Reads a file into the database.
     *
     * @param filename file to read
     * @exception prolog.PrologException Generated upon a parser error.
     * @see prolog.Parser
     */
    public void consult(String filename) throws IOException, PrologException {
	Yylex lex = new Yylex(new BufferedReader(new FileReader(filename)));

	while(!parser.parse(this, lex, false));
    }

    /**
     * Finds the first solution to the given goal. If the goal is
     * <TT>null</TT>, we backtrack and look for another solution.
     */
    public boolean find(Goal goal) throws PrologException {
	if(goal == null) {
	    this.goal = fail;
	} else {
	    this.goal = goal;

	    // HACK
	    runtime_stack_top = 0;
	    cp_last = -1;
	}

// 	System.err.println(this.toString());

	return find();
    }

    /**
     * Executes the given goal once only, and resets the execution
     * environment after doing so.
     */
    public boolean once(Goal goal) throws PrologException {
// 	System.err.println("Prolog.once(): goal: " + goal);

	int cst = cp_stack_top,
	    rtst = runtime_stack_top;
	Goal old_goal = this.goal;

	this.goal = goal;
	boolean success = find();

	this.goal = old_goal;
	cp_stack_top = cst;
	runtime_stack_top = rtst;

	return success;
    }

    /**
     * Does the grunt work - resolution, backtracking.
     *
     * @exception prolog.PrologException Generated by the interpreter on
     * e.g. type exceptions and the parser on syntax errors.
     */
    private boolean find() throws PrologException {
	ChoicePoint cp;

//  	System.err.println("Prolog.find(" + goal + ")");

	// ASSUME GOAL != null

	// Add the goal as the "root" ar
	if((ar_cur = runtime_stack[runtime_stack_top]) == null) {
	    ar_cur = runtime_stack[runtime_stack_top] = new ActivationRecord();
	}

	ar_cur_index     = runtime_stack_top++;
	ar_cur.cont_ar   = -1;
	ar_cur.cut_ptr   = cut_ptr;

// 	System.err.println("Prolog.find(): goal: " + goal);

	while(goal != null) {
//      	    System.err.println("===========\n>> top: runtime_stack_top: " + runtime_stack_top
//  			       + " goal:\n" + goal
//       			       + "\nrs[" + ar_cur_index + "] ar:\n" + ar_cur);

// 	    System.err.println(">> Goal: " + goal);

	    old_cp_stack_top = cp_stack_top;

	    // Attempt to resolve this goal.
	    if(!goal.unify(ar_cur.insts, this)) {
		// Unification failed, backtrack.
		while(true) {
		    if((cp = backtrack()) == null) {
			/* Fixup the runtime stack. FIXME */
			runtime_stack_top = 0;

// 			System.err.println("Prolog.find - fail - goal: " + goal);
			return false;
		    }

		    old_cp_stack_top = cp_stack_top;

		    if(goal.redo(ar_cur.insts, this, cp)) {
			break;
		    }
		}
	    }

	    // Find the next pending goal.
	    while(goal == null) {
		// Pickup the continuation.
		goal    = ar_cur.cont_code;
		cut_ptr = ar_cur.cut_ptr;

		if((ar_cur_index = ar_cur.cont_ar) == -1) {
// 		    System.err.println(">> prolog.find() - success(1).");
		    return true;
		}

		ar_cur  = runtime_stack[ar_cur_index];

// 		System.err.println(">> rs[" + ar_cur_index + "] ar:\n" + ar_cur
// 				   + "\ngoal == null, continue at: " + ar_cur_index
// 				   + " with goal: " + goal
// 				   + " cut_ptr: " + cut_ptr);

		// Pop the runtime stack if possible.
		if(ar_cur_index >= cp_last) {
// 		    System.err.println(" ditching frames > " + ar_cur_index);
		    runtime_stack_top = ar_cur_index + 1;
		}
	    }
	}

	return true;
    }

    /**
     * Allocates a <TT>ChoicePoint</TT>.
     */
    public ChoicePoint allocate_cp(ClauseList next_clause, Binding trail_top) {
//  	System.err.println("Prolog.allocate_cp()");

	ChoicePoint cp;

	if((cp = cp_stack[cp_stack_top]) == null) {
	    cp = cp_stack[cp_stack_top] = new ChoicePoint();
	}

	cp_stack_top++;

	cp.set(ar_cur_index, goal,
	       next_clause,
	       trail_top, cp_last, cut_ptr);

// 	System.err.println("Saving cp[" + (cp_stack_top - 1)
// 			   + "]:\n" + cp_stack[cp_stack_top - 1]);

        // FIXME
	if(cp_last < ar_cur_index) {
// 	    System.err.println(">> Updating cp_last: " + cp_last + " -> " + ar_cur_index + "\n goal: " + goal);
	    cp_last = ar_cur_index;
	}

// 	System.err.println("cp_last': " + cp_last);

	return cp;
    }

    /**
     * Allocates an <TT>ActivationRecord</TT> for the body of a clause.
     *
     * @param new_cut_scope Adjust the <TT>cut_ptr</TT> so that this
     * <TT>Goal</TT> is treated as the body of a unifier of the parent
     * goal, i.e. introduce a new cut scope.
     */
    public void allocate_ar(Term[] insts, Goal body) {
	goal = pending_goal;

	if(body != null) {
	    // Create / reuse an activation record for use by the body
	    // of this goal.
	    if(goal == null
// 	       && old_cp_stack_top == cp_stack_top
	       && ar_cur_index > cp_last) {
		// This ar is completed with the solution of this
		// subgoal, and is not needed on backtracking. LCO it.

// 		System.err.println(">> reusing frame: " + ar_cur_index
// 				   + " body: " + body
// // 				   + " runtime_stack_top: " + runtime_stack_top
// 				   + "\n" + toString());
	    } else {
// 		System.err.println(">> new frame at: " + runtime_stack_top
// 				   + " body: " + body
// // 				   + " continue at: " + ar_cur_index
// 				   + "\n" + toString());

		if((ar_cur = runtime_stack[runtime_stack_top]) == null) {
		    ar_cur = runtime_stack[runtime_stack_top] = new ActivationRecord();
		}

		// When the subgoal is complete, return to the
		// current one.
		ar_cur.cont_ar   = ar_cur_index;
		ar_cur.cont_code = goal;
		ar_cur.cut_ptr   = cut_ptr;

		// Push this ar
		ar_cur_index = runtime_stack_top++;
	    }

	    ar_cur.insts = insts;
	    goal         = body;
	    cut_ptr      = old_cp_stack_top;
	}
    }

    /**
     * Saves a reference to a Variable binding so that it can be undone on
     * backtracking.
     */
    public void push(Binding b) {
	b.next = trail;
	trail = b;
    }

    public void push(Binding b, Term t) {
	b.bind(t);
	b.next = trail;
	trail = b;
    }

    /**
     * Gets the top of the trail, so it can be reset at a later point.
     */
    public Binding get_trail_top() {
	return trail;
    }

    /**
     * Unwinds the trail back to where the argument indicates.
     */
    public void unwind_trail(Binding trail_top) {
	while(trail != trail_top) {
// 	    System.err.println(" >> unbinding: " + trail.getBID());
	    trail.unbind();
	    trail = trail.next;
	}
    }

    /**
     * Backtrack to the last OR-choice we made.
     */
    ChoicePoint backtrack() {
//  	System.err.println(">> backtrack: cp_stack_top: " + cp_stack_top);

	if(cp_stack_top > 0) {
	    ChoicePoint cp = cp_stack[--cp_stack_top];

	    ar_cur_index = cp.cont_ar;
	    ar_cur       = runtime_stack[ar_cur_index];
	    goal         = cp.cont_code;

	    runtime_stack_top = cp_last + 1;
	    cp_last           = cp.cp_last;
	    cut_ptr           = cp.cut_ptr;

	    unwind_trail(cp.trail_top);

//  	    System.err.println(">> backtrack: ar_cur_index: " + ar_cur_index
//  			       + " goal:\n\t" + goal);
// 		System.err.println(">> after backtrack: cut_ptr = " + cut_ptr);

	    return cp;
	} else {
	    return null;
	}
    }

    /**
     * Prunes the <TT>cp_stack</TT> to remove pending non-deterministic
     * goals. Note that the <TT>runtime_stack</TT> is pruned by the
     * <TT>find</TT> method as part of its post-unification cleanup.
     */
    public void cut() {
// 	System.err.println(">> CUT\n==========\n" + toString());

// 	System.err.println(" >> last cp after cut: index = " + (cut_ptr - 1)
// 			   + " cp:\n" + ((cut_ptr - 1 >= 0) ? cp_stack[cut_ptr - 1] : null));
	cp_stack_top = cut_ptr;
// 	old_cp_stack_top = cut_ptr;

	// FIXME
// 	ChoicePoint last_cp_after_cut = cp_stack[cp_stack_top];

// 	if(last_cp_after_cut == null) {
// 	    System.err.println(" last cp after cut doesn't exist! ??");
// 	} else {
// 	    cp_last = last_cp_after_cut.cp_last;
// 	}

//   	System.err.println(toString() + "\n>> ==========");
    }

    public void set_pending_goal(Goal pending_goal) {
	this.pending_goal = pending_goal;
    }

    public void use_pending_goal() {
	this.goal = this.pending_goal;
    }

    /**
     * Convert a list of conjunctions/disjunctions/if-then/if-then-else into
     * a <TT>Goal</TT>.
     * <P>
     * e.g. <TT>','(g1, ','(g2, g3)) -> [g1, g2, g3]</TT>, where the latter
     * is a tree constructed of objects from subclasses of <TT>Goal</TT>.
     * <P>
     * Also, each <TT>Goal</TT> node is linked to the <TT>Predicate</TT>
     * record of the potential unifier for this goal.
     *
     * @param c a list of conjunctions.
     * @return a <TT>Goal</TT> representation of the conjunctions.
     */
    public Goal toGoal(Compound c) {
	return toGoal(c, null);
    }

    private Goal toGoal(Compound c, Goal next) {
//  	System.err.println("toGoal - top - : " + c + " next: " + next);
	String functor = c.functor;
	int arity = c.arity;

	if(functor == Conjunction.FUNCTOR && arity == Conjunction.ARITY) {
	    return toGoal((Compound)c.args[0],
			  toGoal((Compound)c.args[1], next));
	} else if(functor == Disjunction.FUNCTOR && arity == Disjunction.ARITY) {
	    return new Disjunction(toGoal((Compound)c.args[0], next),
				   toGoal((Compound)c.args[1], next));
	} else if(functor == IfThen.FUNCTOR && arity == IfThen.ARITY) {
// 	    System.err.println("toGoal - disj: either: " + either_goal
// 			       + "\nor: " + or_goal
// 			       + "\nnext: " + next);

	    return new IfThen(toGoal((Compound)c.args[0], null),
			      toGoal((Compound)c.args[1], next));
	} else {
	    return new Conjunction(c, next, this);
	}
    }



    /**
     * FIXME: helper function for compiled code.
     */
//     public boolean call(String functor, int arity, Term[] args) throws PrologException {
// 	System.err.println(">> Prolog.call(" + functor + ", " + arity + ")");

// 	for(int i = 0; i < arity; i++) {
// 	    System.err.println("   " + args[i]);
// 	}


// 	Predicate p = db.get(functor, arity);

// 	boolean succ = p.unify(args, this);
//  	System.err.println(">> Prolog.call: " + succ);

// 	return succ;
//     }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("---\n");

	out.append(" ar_cur_index: ");
	out.append(ar_cur_index);
	out.append("\n runtime_stack_top: ");
	out.append(runtime_stack_top);
	out.append("\n goal: ");
	out.append(goal);
	out.append("\n old_cp_stack_top: ");
	out.append(old_cp_stack_top);
	out.append("\n cp_stack_top: ");
	out.append(cp_stack_top);
	out.append("\n cp_last: ");
	out.append(cp_last);
	out.append("\n cut_ptr: ");
	out.append(cp_last);

	out.append("\n---");

	return out.toString();
    }
}
