package prolog;

import java.io.Serializable;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * The information we have about particular <TT>Term</TT> objects: type,
 * type variable assignments, location, argument <TT>TermInfo</TT>s. Not all
 * of this is appropriate in all contexts, but this is a prototype, right?
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@@cse.unsw.edu.au">peteg@@cse.unsw.edu.au</A>)
 * @version $Id: TermInfo.java,v 1.24 2000/11/16 10:30:08 peteg Exp $
 */
public class TermInfo implements Serializable {
    private static int gid = 0;
    private int id = gid++;

    public TermType type;
    public TermInfo[] typevars;
    public TermInfo[] arg_tis;

    /**
     * == <TT>null</TT> if this object has no location (i.e. a
     * destructor/local variable.)
     */
    public Location location;

    /**
     * Linked list stack, for trailing type var's.
     */
    TermInfo next;

    /**
     * true iff the corresponding Term contains unbound variables.
     */
//     boolean contains_variables;

    public TermInfo(TermType type) {
	this.type = type;
    }

    TermInfo(TermType type, TermInfo[] typevars) {
	this.type     = type;
	this.typevars = typevars;
    }

    TermInfo(TermType type, TermInfo[] typevars, TermInfo[] arg_tis) {
	this.type     = type;
	this.typevars = typevars;
	this.arg_tis  = arg_tis;
    }

    TermInfo(TermType type, TermInfo[] typevars, Location location) {
	this.type     = type;
	this.typevars = typevars;
	this.location = location;
    }

    /**
     * @see prolog.ClauseIndexTrad#algebraic_build_index
     */
    void set(TermInfo ti1, Location l) {
// 	System.err.println("TI.set(): this: " + this
// 			   + "\n that: " + ti1 + "\n");
	this.type     = ti1.type;
	this.typevars = ti1.typevars;

	if(type instanceof CompoundType) {
	    // Add some ti's for imaginary arg's, located relative to
	    // this thing.
	    CompoundType component = (CompoundType)type;

	    arg_tis = new TermInfo[component.arity];

	    for(int j = 0; j < component.arity; j++) {
		arg_tis[j] = new TermInfo(component.args[j],
					  typevars,
					  new Location(this, j));
	    }

	    // If this thing has a location, use it, otherwise use the arg.
	    if(ti1.location == null) {
		this.location = l;
	    } else {
		this.location = ti1.location;
	    }
	} else {
	    this.arg_tis  = ti1.arg_tis;
	    this.location = l;
	}
    }

    boolean unify(TermInfo arg, TypeTrail trail) {
// 	boolean succ = type.unify(this, arg, ptd);

// 	try { throw new Exception(); } catch(Exception e) { e.printStackTrace(); }

// 	return succ;

	return type.unify(this, arg, trail);
    }

    /**
     * Loads the <TT>Term</TT> we denote, leaving a reference on the stack.
     * <P>
     * Stack: <TT>... -> ..., TermRef</TT>, where <TT>TermRef</TT> has the
     * type of the loaded/created <TT>Term</TT>.
     *
     * @param ccg the <TT>ClauseCodeGen</TT> context.
     * @param cast should this object be cast to it's most-specific type?
     */
    public void load(ClauseCodeGen ccg, boolean cast) {
	type.load(ccg, this, cast);
    }

    /**
     * Get rid of Locations, etc.
     */
    void clean() {
// 	System.err.println("TI.clean(): cleaning typevars: " + toString());

	if(typevars != null) {
	    for(int i = 0; i < typevars.length; i++) {
		TermInfo ti = typevars[i];

		if(ti != null && ti.location != null) {
		    typevars[i].clean();
		}
	    }
	}

	if(location == null) {
// 	    System.err.println("TI.clean(): stopping for: " + toString());
	} else {
// 	    System.err.println("TI.clean(): cleaning args as well: " + toString());

	    location = null;

	    if(arg_tis != null) {
		for(int i = 0; i < arg_tis.length; i++) {
		    TermInfo ti = arg_tis[i];

		    if(ti != null) {
			ti.clean();
		    }
		}
	    }
	}
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("[TI: ");
	out.append(id);

// 	if(contains_variables) {
// 	    out.append(" not cached");
// 	}

	out.append("]");

	out.append("Type: ");
	type.toString(out);

	if(location != null) {
	    out.append(" Location: ");
	    out.append(location);
	}

	if(arg_tis != null && arg_tis.length > 0) {
	    out.append(" Arg_tis: ");
	    for(int i = 0; i < arg_tis.length; ) {
		TermInfo arg_ti = arg_tis[i];

		out.append(i);
		out.append(": ");

		if(arg_ti == null) {
		    out.append("<null>");
		} else {
		    out.append("[");
		    out.append(arg_ti.toString());
		    out.append("]");
		}

		if(++i < arg_tis.length) {
		    out.append(", ");
		}
	    }
	}

	if(typevars != null && typevars.length > 0) {
	    out.append(" TypeVars: ");
	    for(int i = 0; i < typevars.length; i++) {
		TermInfo typevar = typevars[i];

		out.append("{");
		out.append(i);
		out.append(": ");

		if(typevar == null) {
		    out.append("<null>");
		} else {
		    out.append("[TI: ");
		    out.append(typevar.id);
		    out.append("] ");
		    out.append(typevar.type);
		}

		out.append("} ");
	    }
	}

	return out.toString();
    }
}
