package prolog;

import java.io.Serializable;
import java.util.Enumeration;

import prolog.util.Hashtable;
import prolog.util.Vector;

/**
 * Specialised <TT>Hashtable</TT> implementation: map <TT>(String, int)
 * -> AlgebraicType</TT>.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: TypeDictionary.java,v 1.16 2000/10/18 10:57:51 peteg Exp $
 */
class TD_hashtable implements Serializable {
    private static final int size = 13;
    private FunctorArity[] table;

    TD_hashtable() {
	table = new FunctorArity[size];
    }

    FunctorArity get(int functor_arity) {
	int hash = functor_arity % size;

	for(FunctorArity cur = table[hash]; cur != null; cur = cur.next) {
	    if(cur.functor_arity == functor_arity) {
		return cur;
	    }
	}

	return null;
    }

    void put(int functor_arity,
	     String functor, int arity,
	     AlgebraicType type, CompoundType component) {

	if(get(functor_arity) != null) {
	    throw new RuntimeException("Mapping already defined for "
				       + component);
	}

	FunctorArity cur = new FunctorArity(functor, arity, functor_arity);
	int hash = functor_arity % size;

	cur.next    = table[hash];
	table[hash] = cur;

	cur.type      = type;
	cur.component = component;
    }

    void put(int functor_arity, AlgebraicType type) {
	put(functor_arity, null, -1, type, null);
    }

    void typesToString(StringBuffer out) {
	for(int i = 0; i < size; i++) {
	    for(FunctorArity cur = table[i]; cur != null; cur = cur.next) {
		cur.type.toString(out);
		out.append("\n");
	    }
	}
    }

    void functor_mapToString(StringBuffer out) {
	for(int i = 0; i < size; i++) {
	    for(FunctorArity cur = table[i]; cur != null; cur = cur.next) {
		out.append(cur.functor);
		out.append("/");
		out.append(cur.arity);
		out.append(" -> ");
		out.append(cur.type.toString());
		out.append(" -- ");
		out.append(cur.component.toString());
		out.append("\n");
	    }
	}
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	for(int i = table.length; i-- > 0; ) {
	    for(FunctorArity cur = table[i]; cur != null; cur = cur.next) {
		out.append("bucket: ");
		out.append(i);
		out.append(" ");
		out.append(cur);
		out.append("\n");
	    }
	}

 	return out.toString();
    }
}

/**
 * The dictionary of user- and system-defined types.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: TypeDictionary.java,v 1.16 2000/10/18 10:57:51 peteg Exp $
 * @see prolog.lib._type_def_2
 * @see prolog.lib._has_type_2
 * @see prolog.lib.listtypes_0
 */
public class TypeDictionary implements Serializable {
    /**
     * Maps TypeName (== name/arity) -> AlgebraicType.
     */
    private TD_hashtable types;

    /**
     * Maps SyntacticTerm (== functor/arity) -> TermType.
     */
    private TD_hashtable functor_map;

    TypeDictionary() {
	this.types = new TD_hashtable();
	this.functor_map = new TD_hashtable();
    }

    /**
     * Defines a type with the given name and definition.
     *
     * @see prolog.lib._type_def_2
     */
    public void type_def(Compound c, Compound defs)
	throws ArgumentException {

	/* Maps Binding -> TypeVariable */
	Vector components = new Vector();
	Hashtable vars    = new Hashtable();

	for(int i = 0; i < c.arity; i++) {
	    // The cast is just for safety.
	    vars.put((Binding)c.args[i], new TypeVariable(i, false));
	}

 	AlgebraicType type = new AlgebraicType(c.functor,
					       c.functor_arity,
					       vars.size());
	types.put(c.functor_arity, type);

// 	System.err.println(">> TD: top: types: " + types);

	while(defs != null) {
	    Compound this_def;

	    if(defs.functor == Disjunction.FUNCTOR) {
		this_def = (Compound)defs.args[0];
		defs     = (Compound)defs.args[1];
	    } else {
		this_def = defs;
		defs     = null;
	    }

// 	    System.err.println("TypeDict.type_def(): " + this_def);

	    CompoundType component = new CompoundType(this_def, this, vars);
	    components.addElement(component);

	    /* Map this functor/arity -> the lub type. */
	    FunctorArity elt = functor_map.get(component.functor_arity);

	    if(elt == null) {
		functor_map.put(component.functor_arity,
				component.functor, component.arity,
				type, component);
	    } else {
		throw new RuntimeException("functor already mapped!");
	    }
	}

	type.setComponents(components);

//  	throw new RuntimeException("TypeDictionary.type_def: " + type);
    }

    /**
     * Returns the type record associated with the given
     * <TT>functor/arity</TT> (the name of the type, not it's data
     * constructors).
     * <P>
     * <B>NOTE:</B> This is the original record (i.e. not a copy!)
     *
     * @see prolog.CompoundType
     */
    AlgebraicType get(int functor_arity) {
	FunctorArity elt = types.get(functor_arity);

// 	System.err.println(" Types.get(" + functor_arity + ") -> " + elt);

	return (elt == null) ? null : elt.type;
    }

    /**
     * Determines the type of a function symbol and its arguments. Really
     * only for use by <TT>Compound</TT>.
     *
     * @param map map functors to their most general type.
     * @see prolog.Compound#typeify
     */
    void type_of(Compound c, boolean map) {
// 	System.err.println("TypeDictionary.type_of(" + c + ")");

	FunctorArity elt;

	if(map && (elt = functor_map.get(c.functor_arity)) != null) {
	    // Algebraic type.
	    TermInfo[] typevars = new TermInfo[elt.type.numVars];
	    TermInfo ti = elt.component.instantiate(c, typevars, this, map);

	    // We're mapping, so generalise the type, and toss the details.
	    ti.type    = elt.type;
	    ti.arg_tis = null;
	} else {
	    // Unknown data constructor. Doesn't contains typevar's, then.
	    CompoundType.make(c, this, map);
	}
    }

    public TermType parse_type_sig(Compound c) {
        throw new RuntimeException("TypeDictionary.parse_type_sig: " + c);
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("--- TypeDictionary: known types ---\n");
	types.typesToString(out);

	out.append("\n--- TypeDictionary: functor_map ---\n");
	functor_map.functor_mapToString(out);

	return out.toString();
    }
}
