package prolog.math;

import prolog.*;

/**
 * Maths package for Prolog. A stripped and modified version of the gnu.math
 * library, which has the following copyright:
 * <P>
 * Copyright (c) 1997  Per M.A. Bothner.
 * <B>This is free software;  for terms and warranty disclaimer see ./COPYING.</B>
 * <P>
 * This is the base class of all the other number types.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Numeric.java,v 1.13 2000/10/29 01:29:10 peteg Exp $
 */
public abstract class Numeric extends Term {

/* prolog stuff. */

    public Term instantiate(Term[] insts) {
	return this;
    }

    public boolean unify(Term t, Term[] insts, Prolog p) {
//      System.out.println("Trying to unify: " + this + " with: " + t);

        if(t instanceof Variable) {
            return t.unify(this, insts, p);
	}

        return t instanceof Numeric && equals(t);
    }

    public boolean identical(Term t) {
        return t instanceof Numeric && equals(t);
    }

    public String displayString() {
	return toString();
    }

/* from gnu.math */
    public byte byteValue()    {return (byte)intValue(); }
    public abstract double doubleValue();
    public float floatValue () { return (float)doubleValue(); }
    public int intValue()      { return (int)longValue(); }
    public long longValue()    { return (long)doubleValue(); }
    public short shortValue()  { return (short)intValue(); }

    /** Return this + k * obj. */
//     public abstract Numeric add (Numeric n, int k);

    public abstract Numeric add(Numeric n);
    public abstract Numeric sub(Numeric n);
    public abstract Numeric mul(Numeric n);
    public abstract Numeric div(Numeric n);
    public abstract Numeric mod(Numeric n);

    public abstract Numeric neg();

//     public abstract int sign();
    public abstract boolean isNegative();

    public abstract String toString(int radix);

    public String toString() { return toString (10); }

    /* Rounding modes: */
    public static final int FLOOR = 1;
    public static final int CEILING = 2;
    public static final int TRUNCATE = 3;
    public static final int ROUND = 4;

    /**
     * Return 1 if this&gt;obj; 0 if this==obj; -1 if this&lt;obj; -2 if
     * either is NaN; -3 if not comparable (not equal).
     */
    public abstract int compare(Numeric n);

    public boolean equals(Object obj) {
	if(obj == null || ! (obj instanceof Numeric)) {
	    return false;
	}

	return compare((Numeric)obj) == 0;
    }

    public Numeric abs() {
	return isNegative() ? neg() : this;
    }

    public Numeric toInt(int rounding_mode) {
	double d = doubleValue();
	int i;

	switch(rounding_mode) {
	case FLOOR:
	    i = (int)Math.floor(d);
	    break;
	case CEILING:
	    i = (int)Math.ceil(d);
	    break;
	case TRUNCATE:
	    i = (int)(d < 0.0 ? Math.ceil(d) : Math.floor(d));
	    break;
	case ROUND:
	    i = (int)Math.rint(d);
	    break;
	default:  // Illegal rounding_mode
	    throw new RuntimeException();
	}

	return IntNum.make(i);
    }

    public FloNum exp() {
	return new FloNum(Math.exp(doubleValue()));
    }

  public FloNum log ()
  {
    double x = doubleValue();

    if (x < 0) {
      throw new RuntimeException ();
    }

    return new FloNum(Math.log(x));
  }

    public final FloNum sin() {
	return new FloNum(Math.sin(doubleValue()));
    }

    public final FloNum asin() {
	return new FloNum(Math.asin(doubleValue()));
    }

    public final FloNum sqrt() {
	return new FloNum(Math.sqrt(doubleValue()));
    }

    /**
     * Every <TT>Numeric</TT> subclass must provide a <TT>hashCode</TT>
     * method such that numbers that are equal return the same hash.
     */
    public abstract int hashCode();
}
