/*
 * Approxmation to an ISO Prolog scanner.
 * peteg@cse.unsw.edu.au
 * $Id: prolog.lex,v 1.9 2000/10/02 12:09:33 peteg Exp $
 */

package prolog;

%%

%final
%type void
%line

%{
    /* Constants */

    /* Lexeme types */
    static final int EOF         =  0;

    static final int DOUBLE      =  1;
    static final int INTEGER     =  2;
    static final int STRING      =  3;

    static final int PUNCTUATION =  4;
    static final int CUT         =  5;
    static final int FULL_STOP   =  6;

    static final int VARIABLE    =  7;
    static final int AVARIABLE   =  8;

    static final int WHITESPACE  =  9;

    /* Semantic fields accessible to the parser and error handler. */
    int tokenType;

    int intVal;
    double doubleVal;
    String stringVal;
    char charVal;

    /**
     * EOF has been reached.
     */
    boolean eof = false;

    /**
     * Is the comma (exclusively) punctuation, or could it be a functor?
     */
    private boolean comma_is_punct;

    final int yyline() {
	return yyline + 1;
    }

    private boolean pushed_back = false;

    /**
     * Scans the next token from the <TT>InputStream</TT>.
     *
     * @param Is the comma (exclusively) punctuation, or could it be a functor?
     */
    void scan(boolean comma_is_punct) throws java.io.IOException, prolog.ParsingException {
	if(pushed_back) {
	    pushed_back = false;

	    /* Grotty: check the comma got parsed the right way, if it
               matters. */
	    if(comma_is_punct != this.comma_is_punct) {
		if(tokenType == PUNCTUATION && charVal == ',' && !comma_is_punct) {
/* 		    System.err.println(">> lex: comma lexed as punctuation; changing it to functor."); */
		    tokenType = STRING;
		    stringVal = ",";
		} else if(tokenType == STRING && stringVal.equals(",") && comma_is_punct) {
/* 		    System.err.println(">> lex: comma lexed as functor; changing it to punctuation."); */
		    tokenType = PUNCTUATION;
		    charVal = ',';
		}
	    }
	    return;
	} else {
	    this.comma_is_punct = comma_is_punct;
	    yylex();
	}
    }

    /**
     * Ungets a token. Only allows 1 symbol to be pushed back.
     */
    void unget() {
	if(pushed_back) {
	    throw new RuntimeException("lex.unget(): already pushed_back.");
	} else {
	    pushed_back = true;
/* 	    System.err.println(" lex.unget(): " + this); */
	}
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("line: ");
	out.append(yyline());
	out.append(" ");

	switch(tokenType) {
	case DOUBLE:
	    out.append("double: ");
	    out.append(doubleVal);
	    break;

	case INTEGER:
	    out.append("integer: ");
	    out.append(intVal);
	    break;

	case STRING:
	    out.append("string: ");
	    out.append(stringVal);
	    break;

	case PUNCTUATION:
	    out.append("punctuation: ");
	    out.append(charVal);
	    break;

	case CUT:
	    out.append("cut");
	    break;

	case FULL_STOP:
	    out.append("fullstop");
	    break;

	case VARIABLE:
	    out.append("variable: ");
	    out.append(stringVal);
	    break;

	case AVARIABLE:
	    out.append("avariable: ");
	    out.append(stringVal);
	    break;

	case WHITESPACE:
	    out.append("whitespace");
	    break;

	case EOF:
	    out.append("<<EOF>>");
	    break;

	default:
	    out.append("<< unknown >>");
	    break;
	}

	out.append(" pushed_back: ");
	out.append(pushed_back);

	return out.toString();
    }

    /**
     * Debugging: standalone scanner.
     */
    public static void main(String[] args) {
	try {
	    Yylex lex = new Yylex(System.in);

	    while(!lex.eof) {
		lex.scan(false);
	    }
	} catch(Exception e) {
	    System.err.println(e);
	}
    }

    private int comment_count = 0;
%}

%yylexthrow{
ParsingException
%yylexthrow}

%eof{
    eof = true;
    tokenType = EOF;
/*     System.err.println(">> LEX EOF CODE."); */
%eof} 

%eofval{
    /* JFlex generates a "return null" otherwise. */
/*     System.err.println(">> LEX EOFVAL CODE."); */

/*     Thread.dumpStack(); */
    return;
%eofval}

CommentContent = ( [^*] | \*+[^*/] )*

LineTerminator = \r|\n|\r\n
WhiteSpace     = {LineTerminator} | [ \t\f]

StringText     = (\\\"|[^\n\"]|\\{WhiteSpace}+\\)*

/* Prolog-specific def's. */
Comma          = ","
Cut            = "!"
FullStop       = "\."
Graphic        = ("#"|"\$"|"&"|"*"|"+"|"-"|"."|"/"|":"|"\<"|"="|"\>"|"\?"|"@"|"\\"|"\^"|"~")+
Punctuation    = "("|")"|"\["|"\]"|"\{"|"|"|"\}"
Solo           = ";"|"`"

Functor        = [:lowercase:][:jletterdigit:]*
Variable       = [:uppercase:][:jletterdigit:]*

%% 

/* Normal prolog text. */
<YYINITIAL> {
    /* Strip comments. */
    "%".*{LineTerminator} {
    }

    "/""*"+ [^*] {CommentContent} \*+ "/" {
/* 	System.err.println("-> comment: " + yytext() + " state: " + yystate()); */
    }

    \"{StringText}\" {
	tokenType = STRING;
	stringVal = yytext().substring(1, yytext().length() - 1);
	return;
    }

    \'{StringText}\' {
	tokenType = STRING;
	stringVal = yytext().substring(1, yytext().length() - 1);
	return;
    }

    ("-"|"+")?[:digit:]+"."[:digit:]+ {
	tokenType = DOUBLE;
	doubleVal = (Double.valueOf(yytext())).doubleValue();
/* 	System.err.println("lex: reading DOUBLE: " + doubleVal); */
	return;
    }

    ("-"|"+")?[:digit:]+ {
	tokenType = INTEGER;
	intVal = Integer.parseInt(yytext());
/* 	System.err.println("lex: reading INTEGER: " + intVal); */
	return;
    }

    /* Special case the comma, which is otherwise a Solo character. */
    {Comma} {
	if(comma_is_punct) {
	    tokenType = PUNCTUATION;
	    charVal = ',';
/* 	    System.err.println("lex: comma as punctuation."); */
	} else {
	    tokenType = STRING;
	    stringVal = ",";
/* 	    System.err.println("lex: comma as functor."); */
	}
	return;
    }

    /* Special case the fullstop, which is otherwise a Graphic character. */
    {FullStop}({WhiteSpace}|[#/]) {
	tokenType = FULL_STOP;
/* 	System.err.println("lex: fullstop."); */
	return;
    }

    {Graphic} {
	stringVal = yytext();
	tokenType = STRING;
/* 	System.err.println(" lex graphic: " + stringVal); */
	return;
    }

    {Punctuation} {
	charVal = yytext().charAt(0);
	tokenType = PUNCTUATION;
/* 	System.err.println(" lex punctuation: " + charVal); */
	return;
    }

    /* Special case the cut, which is otherwise a Solo character. */
    {Cut} {
	tokenType = CUT;
/* 	System.err.println("lex: fullstop."); */
	return;
    }

    {Solo} {
	stringVal = yytext();
	tokenType = STRING;
/* 	System.err.println(" lex solo: " + stringVal); */
	return;
    }

    "_" {
	    tokenType = AVARIABLE;
	    return;
    }

    {Functor} {
/* 	System.err.println("lex.scan(): read STRING: " + yytext()); */
	tokenType = STRING;
	stringVal = yytext();
	return;
    }

    {Variable} {
/* 	System.err.println("lex.scan(): read VARIABLE: " + yytext()); */
	tokenType = VARIABLE;
	stringVal = yytext();
	return;
    }

    {WhiteSpace}+ {
/*     tokenType = WHITESPACE;
       return; */
    }

    . {
	throw new ParsingException("Line: " + yyline() + " Illegal character: \"" + yytext() + "\"");
    }
}
