% -*- Prolog -*-

%
% Convert a formula in First Order Logic to clausal (conjunctive normal)
% form.
%
% Appendix B of Clocksin & Mellish.
%
% Caveats: Each variable is introduced by a single quantifier. In other words,
%          variables are renamed so that the order of quantifiers is all that
%          matters.
%
% Existential quantification: exists(Var, formula(Var))
% Universal quantification:   all(Var, formula(Var))
%

% Operators are as follows.
:- op( 30,  fx, ~).	% not
:- op(100, xfy, #).	% or
:- op(100, xfy, &).	% and
:- op(150, xfy, ->).	% implication
:- op(150, xfy, <->).	% bi-implication

test :-
	% "For all x, there exists y such that y = x'"
	translate(all(X, exists(Y, Y = succ(X)))),
	% "There is no x such that x' = 0"
	translate(~(exists(X, succ(X) = 0))),
	% "If x' = y', then x = y"
	translate(all(X, all(Y, succ(X) = succ(Y) -> X = Y))).

% translate/1: Convert a FO formula to clausal form.
translate(X) :-
	implout(X, X1),
	negin(X1, X2),
	skolem(X2, X3, []),
	univout(X3, X4),
	conjn(X4, X5),
	clausify(X5, Clauses, []),
	write([X5, Clauses]), nl,
	pclauses(Clauses).

% Stage 1: Remove implications
implout(P <-> Q, (P1 && Q1) || (~P1 && ~Q1)) :- !,
	implout(P, P1), implout(Q, Q1).
implout(P -> Q, (~P1 || Q1)) :- !,
	implout(P, P1), implout(Q, Q1).
implout(all(X, P), all(X, P1)) :- !,
	implout(P, P1).
implout(exists(X, P), exists(X, P1)) :- !,
	implout(P, P1).
implout((P && Q), (P1 && Q1)) :- !,
	implout(P, P1),
	implout(Q, Q1).
implout((P || Q), (P1 || Q1)) :- !,
	implout(P, P1),
	implout(Q, Q1).
implout((~P), (~P1)) :- !,
	implout(P, P1).
implout(P, P).

% Stage 2: Move negation inwards
negin((~P), P1) :- !, neg(P, P1).
negin(all(X, P), all(X, P1)) :- !, negin(P, P1).
negin(exists(X, P), exists(X, P1)) :- !, negin(P, P1).
negin((P && Q), (P1 && Q1)) :- !, negin(P, P1), negin(Q, Q1).
negin((P || Q), (P1 || Q1)) :- !, negin(P, P1), negin(Q, Q1).
negin(P, P).

neg((~P), P1) :- !, negin(P, P1).
neg(all(X, P), exists(X, P1)) :- !, neg(P, P1).
neg(exists(X, P), all(X, P1)) :- !, neg(P, P1).
neg((P && Q), (P1 || Q1)) :- !, neg(P, P1), neg(Q, Q1).
neg((P || Q), (P1 && Q1)) :- !, neg(P, P1), neg(Q, Q1).
neg(P, (~P)).

% Stage 3: Skolemise (remove existential quantifiers)
skolem(all(X, P), all(X, P1), Vars) :- !, skolem(P, P1, [X|Vars]).
skolem(exists(X, P), P2, Vars) :- !,
	gensym(f, F),
	univ(Sk, [F|Vars]),
	subst(X, Sk, P, P1),
	skolem(P1, P2, Vars).
skolem((P && Q), (P1 && Q1), Vars) :- !,
	skolem(P, P1, Vars), skolem(Q, Q1, Vars).
skolem((P || Q), (P1 || Q1), Vars) :- !,
	skolem(P, P1, Vars), skolem(Q, Q1, Vars).
skolem(P, P, _).

% gensym/2 is implemented in Java.
% subst(X, Sk, F, F1) is true if F1 is the result of substituting Sk for X
%                     in F.
% Note we only have to deal with all/exists/&&/||/~
subst(X, Sk, all(Y, F), all(Y, F1)) :- !,
	subst(X, Sk, F, F1).
subst(X, Sk, exists(Y, F), exists(Y, F1)) :- !,
	subst(X, Sk, F, F1).
subst(X, Sk, (P && Q), (P1 && Q1)) :- !,
	subst(X, Sk, P, P1), subst(X, Sk, Q, Q1).
subst(X, Sk, (P || Q), (P1 || Q1)) :- !,
	subst(X, Sk, P, P1), subst(X, Sk, Q, Q1).
subst(X, Sk, (~P), (~P1)) :- !,
	subst(X, Sk, P, P1).
subst(X, Sk, P, P1) :-
	subst_term(X, Sk, P, P1).

% Substitute Sk for X in an atom.
subst_term(X, Sk, P, P) :- atomic(P).
subst_term(X, Sk, P, Sk) :- identical(X, P), !.
subst_term(X, Sk, P, P1) :-
	univ(P, [Functor|Args]),
	subst_args(X, Sk, Args, Args1),
	univ(P1, [Functor|Args1]).

subst_args(X, Sk, [], []).
subst_args(X, Sk, [Arg|Args], [Sk|Args1]) :-
	identical(X, Arg), !,
	subst_args(X, Sk, Args, Args1).
subst_args(X, Sk, [Arg|Args], [Arg|Args1]) :-
	var(Arg), !,
	subst_args(X, Sk, Args, Args1).
subst_args(X, Sk, [Arg|Args], [Arg1|Args1]) :- !,
	subst_term(X, Sk, Arg, Arg1),
	subst_args(X, Sk, Args, Args1).

% Stage 4: Move universal quantifiers outwards
univout(all(X, P), P1) :- !, univout(P, P1).
univout((P && Q), (P1 && Q1)) :- !,
	univout(P, P1), univout(Q, Q1).
univout((P || Q), (P1 || Q1)) :- !,
	univout(P, P1), univout(Q, Q1).
univout(P, P).

% Stage 5: Distribute "&&" over "||"
conjn((P || Q), R) :- !,
	conjn(P, P1), conjn(Q, Q1),
	conjnl((P1 || Q1), R).
conjn((P && Q), (P1 && Q1)) :- !,
	conjn(P, P1), conjn(Q, Q1).
conjn(P, P).

conjnl(((P && Q) || R), (P1 && Q1)) :- !,
	conjn((P || Q), P1), conjn((Q || R), Q1).
conjnl((P || (Q && R)), (P1 && Q1)) :- !,
	conjn((P || Q), P1), conjn((P || R), Q1).
conjnl(P, P).

% Stage 6: Put into clauses
clausify((P && Q), C1, C2) :- !,
	clausify(P, C1, C3), clausify(Q, C3, C2).
clausify(P, [cl(A, B) | Cs], Cs) :-
	inclause(P, A, [], B, []), !.
clausify(_, C, C).

inclause((P || Q), A, A1, B, B1) :- !,
	inclause(P, A2, A1, B2, B1), inclause(Q, A, A2, B, B2).
inclause((~P), A, A, B1, B) :- !,
	notin(P, A), putin(P, B, B1).
inclause(P, A1, A, B, B) :-
	notin(P, B), putin(P, A, A1).

notin(X,[X|_]) :- !, fail.
notin(X,[_|L]) :- !, notin(X, L).
notin(X, []).

putin(X, [], [X]) :- !.
putin(X, [X|L], L) :- !.
putin(X, [Y|L], [Y|L1]) :- putin(X, L, L1).

% Print out clauses
pclauses([]) :- !, nl, nl.
pclauses([cl(A, B)|Cs]) :- pclause(A, B), nl, pclauses(Cs).
pclause(L, []) :- !, pdisj(L), write(".").
pclause([], L) :- !, write(":- "), pconj(L), write(".").
pclause(L1, L2) :- pdisj(L1), write(" :- "), pconj(L2), write(".").

pdisj([L]) :- !, write(L).
pdisj([L|Ls]) :- write(L), write("; "), pdisj(Ls).

pconj([L]) :- !, write(L).
pconj([L|Ls]) :- write(L), write(", "), pconj(Ls).
