% -*- Prolog -*-

% SIMPLE DEMONSTRATION OF GENETIC PROGRAMMING (GP) IN PROLOG.
% (Note that some predicates used may be specific to SEPIA Prolog.)
% Concerning GP see Koza J., `Genetic Programming,' MIT Press, 1992.

% Ismail Ikram  (g93i0527@cs.ru.ac.za), Rhodes University, 1995.
% http://cs.ru.ac.za/homes/g93i0527/

% USAGE --------------------------------------------------------------

% Runs GP when given arguments: population size N, depth limit for
% population members, followed by the function, terminal and training
% sets.
gp(N, Depth, FSet, TSet, TestSet) :-
	create_pop(N, Depth, FSet, TSet, Pop),
	evolve_pop(Pop, Depth, FSet, TSet, TestSet).

% Thus to try out the pi*x*x regression problem as specified in the
% PROBLEM-DEPENDENT section below, give the query:
%         fset(F), tset(T), pi_x_squared(P), gp(10, 5, F, T, P).
% Doing so should display a continuous trace of (hopefully improving)
% fitness values as GP attempts to refine a random initial
% population of 10 members.

% PROBLEM-DEPENDENT -------------------------------------------------

% Function set should be a list of pairs [(operator1, arity1),
% (operator2, arity2), ...] where each operator has been predefined.
fset([(+, 2), (-, 2), (*, 2), (sdiv, 2)]).

% `Safe' division.
sdiv(_, Zero, 0) :-
	Zero =:= 0, !.
sdiv(X, Y, Z) :-
	Z is X/Y.

% Terminal set should be a list of symbols, e.g. [constant1,
% constant2, ..., variable1, variable2, ...]. Variable names may be
% repeated to give them a higher chance of being picked when creating
% random expression trees
tset([1, 2, 3, x, x, x]).

% The training set should be a list of training cases. For terminal
% sets with V variables, the training cases should be a list of V+1
% items: 
%        [desired, (v1, v1val), (v2, v2val), ..., (vV, vVval)] 
% where the first item is the desired output for the training case and
% the remaining items are (variable_name, value) pairs for each
% variable.

% Training set for pi*x*x regression problem.
pi_x_squared([[28.26, (x, 3)], [50.24, (x, 4)],
	      [3.14, (x, 1)], [12.56, (x, 2)], 
	      [0.0, (x, 0)],
	      [3.14, (x, -1)], [12.56, (x, -2)], 
	      [28.26, (x, -3)], [50.24, (x, -4)]]).

% PREDICATES ---------------------------------------------------------

% These predicates are defined in the remaining sections:

% - INITIALIZATION	     - EVOLUTION          
% create_ft		     evolve_pop           
% create_pop		     write_current_best   
% 			     tourn_select         
% - SUBTREE OPERATIONS	     best_select          
% get_subtree		     ran_replace          
% replace_subtree	     worst_replace        
% mutate			                          
% xover			     - UTILITIES          
% 			     ran                  
% - EVALUATION		     getmem               
% eval			     pos_of               
% fitness		     one_upto             
% abs_error		     get_arg              
% fit_pop		     sum_list             
% 			     max_list             
% 			     min_list             

% INITIALIZATION -----------------------------------------------------

% Creates full tree of given depth. Nonterminal nodes contain members
% of FSet and leaf nodes contain members of TSet.
create_ft(0, _, TSet, T) :-
	length(TSet, LT),
	ran(LT, Pos),
	getmem(Pos, TSet, T), !.
create_ft(N, FSet, TSet, [Operator|Args]) :-
	length(FSet, FL),
	ran(FL, Pos),
	getmem(Pos, FSet, (Operator, Arity)),
	NMinus1 is N-1, 
	create_pop(Arity, NMinus1, FSet, TSet, Args).

% Creates list of N full trees having given depth, and function and
% terminal sets.
create_pop(1, Depth, FSet, TSet, [Subtree]) :-
	create_ft(Depth, FSet, TSet, Subtree), !.
create_pop(N, Depth, FSet, TSet, [First|Rest]) :-
	create_ft(Depth, FSet, TSet, First),
	NMinus1 is N-1, 
	create_pop(NMinus1, Depth, FSet, TSet, Rest).

% SUBTREE OPERATIONS ------------------------------------------------

% Gets a random subtree of the given tree at the given depth, if
% possible. Also returns a list of paths taken from the root in order
% to arrive at the subtree (in reverse order).
get_subtree(Tree, 0, [], Tree) :- !.
get_subtree(Terminal, _, [], Terminal) :- atomic(Terminal), !.
get_subtree(Tree, MaxDepth, PathList, Subtree) :-
	TreeTerm =.. Tree,
	functor(TreeTerm, _, Arity),
	ran(Arity, Ran),
	First is Ran+1,
	arg(First, TreeTerm, Arg),
	MaxMinus1 is MaxDepth-1,
	get_subtree(Arg, MaxMinus1, Rest, Subtree),
	append(Rest, [First], PathList).

% Replace a subtree. Arguments:
% 1. Path in given tree from root to the root of subtree to be
% replaced. 
% 2. Subtree to insert into the tree.
% 3. A pair (Tree, Path) where Tree is the given tree and Path is the
% path taken to arrive at Tree's root (initially this is just []).
% 4. The result of inserting the given subtree into the given tree at
% the specified position.
replace_subtree([], Subtree, (_, _), Subtree) :- !.
replace_subtree(TargetPath, Subtree, (_, TargetPath), Subtree) :- !.
replace_subtree(_, _, (Terminal, _), Terminal) :-
	atomic(Terminal), !.
replace_subtree(TargetPath, Subtree, (Tree, CurrentPath), [Op|ResultArgs]) :-
	TreeTerm =.. Tree,
	functor(TreeTerm, Op, Arity),
	one_upto(Arity, CurrentPath, SubtreePaths),
	maplist(get_arg(TreeTerm), SubtreePaths, TreeArgs),
	maplist(replace_subtree(TargetPath, Subtree), 
                TreeArgs, ResultArgs).
    
mutate(Before, Depth, FSet, TSet, After) :-
	DPlus1 is Depth+1,
	ran(DPlus1, MaxDepthOfTarget),
	get_subtree(Before, MaxDepthOfTarget, PathToTarget, _),
	create_ft(Depth, FSet, TSet, Mutant),
	replace_subtree(PathToTarget, Mutant, (Before, []), After).

xover(P1, P2, Depth, C1, C2) :-
	DPlus1 is Depth+1,
	ran(DPlus1, MaxDepthOfTarget),
	get_subtree(P1, MaxDepthOfTarget, Path1, Subtree1),
	get_subtree(P2, MaxDepthOfTarget, Path2, Subtree2),
	replace_subtree(Path1, Subtree2, (P1, []), C1),
	replace_subtree(Path2, Subtree1, (P2, []), C2).

% EVALUATION --------------------------------------------------------

% Evaluate a tree using given variable-value mapping.
eval(Vars, Tree, Value) :- 
	atomic(Tree),
	member((Tree, Value), Vars), !.		
eval(_, Tree, Tree) :- 
	atomic(Tree), !.			
eval(Vars, [Operator|Args], Result) :-
	maplist(eval(Vars), Args, EvaluatedArgs),
	Goal =.. [Operator|EvaluatedArgs],
	Result is Goal.

% Compute fitness as negative of sum of absolute errors over all
% test cases. 
fitness(TestSet, Tree, Fit) :- 
	maplist(abs_error(Tree), TestSet, ErrorList),
	sum_list(ErrorList, SumOfErrors),
	Fit is -SumOfErrors.
abs_error(Tree, [Desired|VarList], Error) :-
	eval(VarList, Tree, Result),
	Diff is Desired - Result,
	Error is abs(Diff).

% Compute the fitness of all members of the given population, using
% given training set.
fit_pop(TestSet, Pop, FitList) :-
	maplist(fitness(TestSet), Pop, FitList).

% EVOLUTION ---------------------------------------------------------

% Applies genetic operators to a population indefinitely, displaying
% the current fittest member after each crossover. Runs a steady-state
% genetic algorithm where one parent is the best member, while the
% other is picked using tournament selection. All children are mutated
% before being placed into the population, at the expense of the worst
% member and another randomly selected member. Arguments: the
% population as a list, maximum depth allowed for expression trees,
% followed by the function, terminal and training sets. 
evolve_pop(Pop, Depth, FSet, TSet, TestSet) :-
	fit_pop(TestSet, Pop, FitList),		% Evaluates all even
						% though there
						% only 2 new members!
	write_current_best(Pop, FitList),
	best_select(Pop, FitList, P1),
	tourn_select(Pop, FitList, P2),
	ran(Depth, MaxXoverDepth),		
	xover(P1, P2, MaxXoverDepth, C1, C2),
	ran(Depth, MaxMutDepth),		
	mutate(C1, MaxMutDepth, FSet, TSet, Child1),
	mutate(C2, MaxMutDepth, FSet, TSet, Child2),
	worst_replace(Pop, FitList, Child1, IntermediatePop),
	ran_replace(IntermediatePop, Child2, NewPop),
	!,					% Save some space.
	evolve_pop(NewPop, Depth, FSet, TSet, TestSet).

% Display the best member in the given population list, and its
% fitness. The given fitness list should specify the current fitness
% of each member. Assumes that the best member has the greatest
% fitness value.
write_current_best(Pop, FitList) :-
	max_list(FitList, MaxFit),
	pos_of(MaxFit, FitList, Pos),
	getmem(Pos, Pop, Tree),
	write("\n\nBest = "), 
	write(MaxFit),
	write("\n"), 
	write(Tree).

% Tournament-select a member of the given population. The given
% fitness list should specify the current fitness of each
% member. Tournament size is 2, and sampling is done with
% replacement.
tourn_select(Pop, FitList, Tree) :-
	length(Pop, N),
	ran(N, Candidate1), 
	ran(N, Candidate2),
	getmem(Candidate1, FitList, Fit1),
	getmem(Candidate2, FitList, Fit2),
	(Fit1 >= Fit2, Winner = Candidate1, !; 
	    Winner = Candidate2),
	getmem(Winner, Pop, Tree).

% Select a member having the greatest fitness.
best_select(Pop, FitList, Tree) :-
	max_list(FitList, MaxFit),
	pos_of(MaxFit, FitList, Winner),
	getmem(Winner, Pop, Tree).

% Replace a randomly selected member of the given population with the
% given expression tree.
ran_replace(Pop, Tree, NewPop) :-
	length(Pop, N),
	ran(N, ReplacePos),
	append(Head, [_|Tail], Pop),
	length(Head, ReplacePos),
	append(Head, [Tree|Tail], NewPop).
  
% Replace a member having the lowest fitness with the given tree.
worst_replace(Pop, FitList, Tree, NewPop) :-
	min_list(FitList, MinFit),
	pos_of(MinFit, FitList, ReplacePos),
	append(Head, [_|Tail], Pop),
	length(Head, ReplacePos),
	append(Head, [Tree|Tail], NewPop).

% UTILITIES ----------------------------------------------------------

% Returns random integer in range [0..Upper).
ran(Upper, X) :-				
	random(R), X is R mod Upper.

% Returns the (N+1)-th member of the given list.
getmem(0, [X|_], X) :- !.
getmem(N, [_|Tail], X) :-
	NMinus1 is N-1,
	getmem(NMinus1, Tail, X).

% Position of the given element in the list, counting from 0 up. 0 if
% Elt not member of List.
pos_of(Elt, List, Pos) :- 
	append(Head, [Elt|_], List), 
	length(Head, Pos), !.
pos_of(_,_,0).

% Generate a list of N lists whose head elements are in the range
% [1..N]. Append the given postfix to each head element.
one_upto(1, Postfix, [[1|Postfix]]) :- !.
one_upto(N, Postfix, List) :-
	NMinus1 is N-1,
	one_upto(NMinus1, Postfix, SmallList),
	append(SmallList, [[N|Postfix]], List).

% Get the ArgNum-th argument of Term, paired with the given list.
get_arg(Term, [ArgNum|Rest], (Arg, [ArgNum|Rest])) :-
	arg(ArgNum, Term, Arg).

% Sum of all numbers in the given list.
sum_list([], 0).
sum_list([H|T], Result) :-
	sum_list(T, R),
	Result is H + R.

% Maximum of the given list. 
max_list(List, Max) :-
	sort(List, Sorted), 
	append(_, [Max], Sorted).

% Minimum of the given list. 
min_list(List, Min) :-
	sort(List, [Min|_]).
